import json
from datetime import datetime

import boto3
from dateutil.tz import tzlocal

from moto import mock_aws

TEST_REGION = "us-west-1"


@mock_aws
def test_create_thing_with_simple_cloudformation():
    # given
    stack_name = "test_stack"
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Thing CloudFormation",
        "Resources": {
            "testThing": {
                "Type": "AWS::IoT::Thing",
            },
        },
        "Outputs": {
            "ThingArn": {"Value": {"Fn::GetAtt": ["testThing", "Arn"]}},
            "ThingId": {"Value": {"Fn::GetAtt": ["testThing", "Id"]}},
        },
    }

    # when
    template_json = json.dumps(template)
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=template_json)

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    things = iot_conn.list_things()["things"]
    assert len(things) == 1
    assert things[0]["thingName"].startswith("test_stack-testThing-")
    assert things[0]["thingArn"].startswith(
        "arn:aws:iot:us-west-1:123456789012:thing/test_stack-testThing-"
    )
    assert things[0]["attributes"] == {}


@mock_aws
def test_create_thing_with_attributes_through_cloudformation():
    # given
    test_thing_name = "Test_Thing"
    stack_name = "test_stack"

    thing_attributes = {"attributes": {"attr1": "value1", "attr2": "value2"}}
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Thing CloudFormation",
        "Resources": {
            "testThing": {
                "Type": "AWS::IoT::Thing",
                "Properties": {
                    "ThingName": test_thing_name,
                    "AttributePayload": json.dumps(thing_attributes),
                },
            },
        },
        "Outputs": {
            "ThingArn": {"Value": {"Fn::GetAtt": ["testThing", "Arn"]}},
            "ThingId": {"Value": {"Fn::GetAtt": ["testThing", "Id"]}},
        },
    }

    # when
    template_json = json.dumps(template)
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=template_json)

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_things()["things"]) == 1

    # describe thing
    resp = iot_conn.describe_thing(thingName=test_thing_name)
    assert resp["thingName"] == test_thing_name
    assert resp["thingArn"] == "arn:aws:iot:us-west-1:123456789012:thing/Test_Thing"
    assert resp["attributes"] == thing_attributes["attributes"]

    # Check stack outputs
    stack = cfn_conn.describe_stacks(StackName=stack_name)["Stacks"][0]
    outputs = {
        Output["OutputKey"]: Output["OutputValue"] for Output in stack["Outputs"]
    }
    assert outputs["ThingArn"] == resp["thingArn"]
    assert outputs["ThingId"] == resp["thingId"]


@mock_aws
def test_update_thing_name_should_recreate_thing_and_change_id():
    # given
    test_thing_name = "Test_Thing"
    updated_thing_name = "Updated_Thing"
    stack_name = "test_stack"

    first_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Thing CloudFormation",
        "Resources": {
            "testThing": {
                "Type": "AWS::IoT::Thing",
                "Properties": {"ThingName": test_thing_name},
            },
        },
    }
    updated_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Thing CloudFormation",
        "Resources": {
            "testThing": {
                "Type": "AWS::IoT::Thing",
                "Properties": {"ThingName": updated_thing_name},
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(first_template))

    # then check list of things and the only one thing
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_things()["things"]) == 1
    first_resp = iot_conn.describe_thing(thingName=test_thing_name)
    assert first_resp["thingName"] == test_thing_name
    assert (
        first_resp["thingArn"] == "arn:aws:iot:us-west-1:123456789012:thing/Test_Thing"
    )

    # and
    cfn_conn.update_stack(
        StackName=stack_name, TemplateBody=json.dumps(updated_template)
    )
    assert len(iot_conn.list_things()["things"]) == 1
    updated_resp = iot_conn.describe_thing(thingName=updated_thing_name)
    assert updated_resp["thingName"] == updated_thing_name
    assert (
        updated_resp["thingArn"]
        == "arn:aws:iot:us-west-1:123456789012:thing/Updated_Thing"
    )

    # and it's a different thing
    assert first_resp["thingId"] != updated_resp["thingId"]


@mock_aws
def test_delete_resource_should_delete_thing():
    # given
    test_thing_name = "Test_Thing"
    stack_name = "test_stack"

    first_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Thing CloudFormation",
        "Resources": {
            "testThing": {
                "Type": "AWS::IoT::Thing",
                "Properties": {
                    "ThingName": test_thing_name,
                },
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(first_template))

    # then check list of things and the only one thing
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_things()["things"]) == 1

    # and
    cfn_conn.delete_stack(StackName=stack_name)
    assert len(iot_conn.list_things()["things"]) == 0


@mock_aws
def test_create_thingtype_with_simple_cloudformation():
    # given
    stack_name = "test_stack"
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT ThingType CloudFormation",
        "Resources": {
            "testThingType": {
                "Type": "AWS::IoT::ThingType",
            },
        },
        "Outputs": {
            "ThingTypeArn": {"Value": {"Fn::GetAtt": ["testThingType", "Arn"]}},
            "ThingTypeId": {"Value": {"Fn::GetAtt": ["testThingType", "Id"]}},
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(template))

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    thing_types = iot_conn.list_thing_types()["thingTypes"]
    assert len(thing_types) == 1
    assert thing_types[0]["thingTypeName"].startswith("test_stack-testThingType-")
    assert thing_types[0]["thingTypeArn"].startswith(
        "arn:aws:iot:us-west-1:123456789012:thingtype/test_stack-testThingType-"
    )

    # and check if they are the same as in a stack outputs
    stack = cfn_conn.describe_stacks(StackName=stack_name)["Stacks"][0]
    outputs = {
        Output["OutputKey"]: Output["OutputValue"] for Output in stack["Outputs"]
    }
    assert outputs["ThingTypeArn"] == thing_types[0]["thingTypeArn"]


@mock_aws
def test_update_thingtype_using_cloudformation_should_recreate_id():
    # given
    stack_name = "test_stack"
    initial_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT ThingType CloudFormation",
        "Resources": {
            "testThingType": {
                "Type": "AWS::IoT::ThingType",
                "Properties": {"ThingTypeName": "TestThing1"},
            },
        },
    }

    updated_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT ThingType CloudFormation",
        "Resources": {
            "testThingType": {
                "Type": "AWS::IoT::ThingType",
                "Properties": {"ThingTypeName": "TestThing2"},
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(
        StackName=stack_name, TemplateBody=json.dumps(initial_template)
    )

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    initial_thing_types = iot_conn.list_thing_types()["thingTypes"]
    assert len(initial_thing_types) == 1
    assert initial_thing_types[0]["thingTypeName"] == "TestThing1"

    # then update
    cfn_conn.update_stack(
        StackName=stack_name, TemplateBody=json.dumps(updated_template)
    )
    updated_thing_types = iot_conn.list_thing_types()["thingTypes"]
    assert len(updated_thing_types) == 1
    assert updated_thing_types[0]["thingTypeName"] == "TestThing2"

    # and ids are different
    assert (
        initial_thing_types[0]["thingTypeArn"] != updated_thing_types[0]["thingTypeArn"]
    )


@mock_aws
def test_cloudformation_stack_delete_should_remove_thingtype():
    # given
    stack_name = "test_stack"
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT ThingType CloudFormation",
        "Resources": {
            "testThingType": {
                "Type": "AWS::IoT::ThingType",
                "Properties": {"ThingTypeName": "TestThingType"},
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(template))

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_thing_types()["thingTypes"]) == 1

    # then update
    cfn_conn.delete_stack(StackName=stack_name)
    assert len(iot_conn.list_thing_types()["thingTypes"]) == 0


@mock_aws
def test_create_policy_with_simple_cloudformation():
    # given
    stack_name = "test_stack"
    policy_name = "Test_Policy"
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Policy CloudFormation",
        "Resources": {
            "testPolicy": {
                "Type": "AWS::IoT::Policy",
                "Properties": {
                    "PolicyName": policy_name,
                    "PolicyDocument": json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": ["iot:Connect"],
                                    "Resource": [
                                        "arn:aws:iot:us-east-1:123456789012:client/client1"
                                    ],
                                }
                            ],
                        }
                    ),
                },
            },
        },
        "Outputs": {
            "PolicyArn": {"Value": {"Fn::GetAtt": ["testPolicy", "Arn"]}},
            "PolicyId": {"Value": {"Fn::GetAtt": ["testPolicy", "Id"]}},
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(template))

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    policies = iot_conn.list_policies()["policies"]
    assert len(policies) == 1
    assert policies[0]["policyName"] == policy_name
    assert (
        policies[0]["policyArn"]
        == "arn:aws:iot:us-west-1:123456789012:policy/Test_Policy"
    )

    # then check stack
    stack = cfn_conn.describe_stacks(StackName=stack_name)["Stacks"][0]
    outputs = {
        Output["OutputKey"]: Output["OutputValue"] for Output in stack["Outputs"]
    }
    assert outputs["PolicyArn"] == policies[0]["policyArn"]


@mock_aws
def test_update_policy_with_the_same_name_should_update_policy_document():
    # given
    stack_name = "test_stack"
    policy_name = "Test_Policy"
    initial_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Policy CloudFormation",
        "Resources": {
            "testPolicy": {
                "Type": "AWS::IoT::Policy",
                "Properties": {
                    "PolicyName": policy_name,
                    "PolicyDocument": json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": ["iot:Connect"],
                                    "Resource": [
                                        "arn:aws:iot:us-east-1:123456789012:client/client1"
                                    ],
                                }
                            ],
                        }
                    ),
                },
            },
        },
    }
    updated_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Policy CloudFormation",
        "Resources": {
            "testPolicy": {
                "Type": "AWS::IoT::Policy",
                "Properties": {
                    "PolicyName": policy_name,
                    "PolicyDocument": json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": ["iot:Connect", "iot:Subscribe"],
                                    "Resource": ["*"],
                                }
                            ],
                        }
                    ),
                },
            },
        },
    }
    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(
        StackName=stack_name, TemplateBody=json.dumps(initial_template)
    )

    # then update stack
    cfn_conn.update_stack(
        StackName=stack_name, TemplateBody=json.dumps(updated_template)
    )

    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    policy = iot_conn.get_policy(policyName=policy_name)

    assert policy["policyName"] == policy_name
    policyDocument = json.loads(policy["policyDocument"])
    assert policyDocument["Statement"][0]["Resource"] == ["*"]


@mock_aws
def test_update_policy_with_different_name_should_recreate_whole_policy():
    # given
    stack_name = "test_stack"
    initial_policy_name = "Test_Policy"
    updated_policy_name = "New_Test_Policy"
    initial_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Policy CloudFormation",
        "Resources": {
            "testPolicy": {
                "Type": "AWS::IoT::Policy",
                "Properties": {
                    "PolicyName": initial_policy_name,
                    "PolicyDocument": json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": ["iot:Connect"],
                                    "Resource": [
                                        "arn:aws:iot:us-east-1:123456789012:client/client1"
                                    ],
                                }
                            ],
                        }
                    ),
                },
            },
        },
    }
    updated_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Policy CloudFormation",
        "Resources": {
            "testPolicy": {
                "Type": "AWS::IoT::Policy",
                "Properties": {
                    "PolicyName": updated_policy_name,
                    "PolicyDocument": json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": ["iot:Connect", "iot:Subscribe"],
                                    "Resource": ["*"],
                                }
                            ],
                        }
                    ),
                },
            },
        },
    }
    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(
        StackName=stack_name, TemplateBody=json.dumps(initial_template)
    )
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    policies = iot_conn.list_policies()["policies"]
    assert len(policies) == 1
    assert policies[0]["policyName"] == initial_policy_name

    # then update stack
    cfn_conn.update_stack(
        StackName=stack_name, TemplateBody=json.dumps(updated_template)
    )

    policies = iot_conn.list_policies()["policies"]
    assert len(policies) == 1
    assert policies[0]["policyName"] == updated_policy_name

    policy = iot_conn.get_policy(policyName=updated_policy_name)

    assert policy["policyName"] == updated_policy_name
    policyDocument = json.loads(policy["policyDocument"])
    assert policyDocument["Statement"][0]["Resource"] == ["*"]


@mock_aws
def test_delete_stack_should_delete_policy():
    # given
    stack_name = "test_stack"
    policy_name = "Test_Policy"
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT Policy CloudFormation",
        "Resources": {
            "testPolicy": {
                "Type": "AWS::IoT::Policy",
                "Properties": {
                    "PolicyName": policy_name,
                    "PolicyDocument": json.dumps(
                        {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": ["iot:Connect"],
                                    "Resource": [
                                        "arn:aws:iot:us-east-1:123456789012:client/client1"
                                    ],
                                }
                            ],
                        }
                    ),
                },
            },
        },
    }
    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(template))
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_policies()["policies"]) == 1

    # then delete stack
    cfn_conn.delete_stack(StackName=stack_name)
    assert len(iot_conn.list_policies()["policies"]) == 0


@mock_aws
def test_create_role_alias_with_simple_cloudformation():
    # given
    stack_name = "test_stack"
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT RoleAlias CloudFormation",
        "Resources": {
            "testRoleAlias": {
                "Type": "AWS::IoT::RoleAlias",
                "Properties": {
                    "RoleAlias": "TestRoleAlias",
                    "RoleArn": "arn:aws:iam::123456789012:role/my-role",
                },
            },
        },
        "Outputs": {
            "RoleAliasArn": {
                "Value": {"Fn::GetAtt": ["testRoleAlias", "RoleAliasArn"]}
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(template))

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert iot_conn.list_role_aliases()["roleAliases"] == ["TestRoleAlias"]

    # then check stack
    stack = cfn_conn.describe_stacks(StackName=stack_name)["Stacks"][0]
    outputs = {
        Output["OutputKey"]: Output["OutputValue"] for Output in stack["Outputs"]
    }
    assert (
        outputs["RoleAliasArn"]
        == "arn:aws:iot:us-west-1:123456789012:rolealias/TestRoleAlias"
    )


@mock_aws
def test_update_role_alias_with_cloudformation():
    # given
    stack_name = "test_stack"
    initial_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT RoleAlias CloudFormation",
        "Resources": {
            "testRoleAlias": {
                "Type": "AWS::IoT::RoleAlias",
                "Properties": {"RoleArn": "arn:aws:iam::123456789012:role/my-role"},
            },
        },
    }
    updated_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT RoleAlias CloudFormation",
        "Resources": {
            "testRoleAlias": {
                "Type": "AWS::IoT::RoleAlias",
                "Properties": {
                    "RoleAlias": "TestRoleAlias",
                    "RoleArn": "arn:aws:iam::123456789012:role/other-role",
                    "CredentialDurationSeconds": 1234,
                },
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(
        StackName=stack_name, TemplateBody=json.dumps(initial_template)
    )

    # then check list of role aliases
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    role_aliases = iot_conn.list_role_aliases()["roleAliases"]
    assert len(role_aliases) == 1
    assert role_aliases[0].startswith("test_stack-testRoleAlias-")

    # then check stack
    cfn_conn.update_stack(
        StackName=stack_name, TemplateBody=json.dumps(updated_template)
    )
    assert iot_conn.list_role_aliases()["roleAliases"] == ["TestRoleAlias"]

    # and describe that role alias
    role_alias = iot_conn.describe_role_alias(roleAlias="TestRoleAlias")
    assert role_alias["roleAliasDescription"]["roleAlias"] == "TestRoleAlias"
    assert (
        role_alias["roleAliasDescription"]["roleAliasArn"]
        == "arn:aws:iot:us-west-1:123456789012:rolealias/TestRoleAlias"
    )
    assert (
        role_alias["roleAliasDescription"]["roleArn"]
        == "arn:aws:iam::123456789012:role/other-role"
    )


@mock_aws
def test_delete_role_alias_with_cloudformation():
    # given
    stack_name = "test_stack"
    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT RoleAlias CloudFormation",
        "Resources": {
            "testRoleAlias": {
                "Type": "AWS::IoT::RoleAlias",
                "Properties": {
                    "RoleAlias": "TestRoleAlias",
                    "RoleArn": "arn:aws:iam::123456789012:role/my-role",
                },
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(template))

    # then check list of role aliases
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_role_aliases()["roleAliases"]) == 1

    # then check stack
    cfn_conn.delete_stack(StackName=stack_name)
    assert iot_conn.list_role_aliases()["roleAliases"] == []


@mock_aws
def test_create_job_template_with_simple_cloudformation():
    # given
    stack_name = "test_stack"
    job_document = {"field": "value"}

    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT JobTemplate CloudFormation",
        "Resources": {
            "testJobTemplate": {
                "Type": "AWS::IoT::JobTemplate",
                "Properties": {
                    "JobTemplateId": "JobTemplate",
                    "Description": "Job template Description",
                    "Document": json.dumps(job_document),
                    "DocumentSource": "a document source link",
                    "PresignedUrlConfig": {
                        "ExpiresInSec": 123,
                        "RoleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
                    },
                    "TimeoutConfig": {"InProgressTimeoutInMinutes": 30},
                },
            },
        },
        "Outputs": {
            "JobTemplateArn": {"Value": {"Fn::GetAtt": ["testJobTemplate", "Arn"]}},
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(template))

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_job_templates()["jobTemplates"]) == 1
    assert (
        iot_conn.list_job_templates()["jobTemplates"][0]["jobTemplateId"]
        == "JobTemplate"
    )

    # then check stack
    stack = cfn_conn.describe_stacks(StackName=stack_name)["Stacks"][0]
    outputs = {
        Output["OutputKey"]: Output["OutputValue"] for Output in stack["Outputs"]
    }
    assert (
        outputs["JobTemplateArn"]
        == "arn:aws:iot:us-west-1:123456789012:jobtemplate/JobTemplate"
    )

    # and describe it
    job_template = iot_conn.describe_job_template(jobTemplateId="JobTemplate")
    assert job_template["jobTemplateId"] == "JobTemplate"
    assert (
        job_template["jobTemplateArn"]
        == "arn:aws:iot:us-west-1:123456789012:jobtemplate/JobTemplate"
    )
    assert job_template["description"] == "Job template Description"
    assert job_template["document"] == '{"field": "value"}'
    assert job_template["documentSource"] == "a document source link"
    assert job_template["createdAt"] == datetime(2015, 1, 1, 0, 0, tzinfo=tzlocal())
    assert job_template["presignedUrlConfig"] == {
        "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
        "expiresInSec": 123,
    }
    assert job_template["timeoutConfig"] == {
        "inProgressTimeoutInMinutes": 30,
    }


@mock_aws
def test_update_job_template_with_simple_cloudformation():
    # given
    stack_name = "test_stack"

    initial_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT JobTemplate CloudFormation",
        "Resources": {
            "testJobTemplate": {
                "Type": "AWS::IoT::JobTemplate",
                "Properties": {
                    "JobTemplateId": "JobTemplate",
                    "Description": "Job template Description",
                    "Document": json.dumps({"field1": "value1"}),
                },
            },
        },
    }
    updated_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT JobTemplate CloudFormation",
        "Resources": {
            "testJobTemplate": {
                "Type": "AWS::IoT::JobTemplate",
                "Properties": {
                    "JobTemplateId": "JobTemplate2",
                    "Description": "Job template Description",
                    "Document": json.dumps({"field2": "value2"}),
                },
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(
        StackName=stack_name, TemplateBody=json.dumps(initial_template)
    )

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_job_templates()["jobTemplates"]) == 1
    assert (
        iot_conn.list_job_templates()["jobTemplates"][0]["jobTemplateId"]
        == "JobTemplate"
    )

    # then update stack
    cfn_conn.update_stack(
        StackName=stack_name, TemplateBody=json.dumps(updated_template)
    )
    assert len(iot_conn.list_job_templates()["jobTemplates"]) == 1
    assert (
        iot_conn.list_job_templates()["jobTemplates"][0]["jobTemplateId"]
        == "JobTemplate2"
    )


@mock_aws
def test_delete_job_template_with_simple_cloudformation():
    # given
    stack_name = "test_stack"

    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "IOT JobTemplate CloudFormation",
        "Resources": {
            "testJobTemplate": {
                "Type": "AWS::IoT::JobTemplate",
                "Properties": {
                    "JobTemplateId": "JobTemplate",
                    "Description": "Job template Description",
                    "Document": json.dumps({"field": "value"}),
                },
            },
        },
    }

    # when
    cfn_conn = boto3.client("cloudformation", region_name=TEST_REGION)
    cfn_conn.create_stack(StackName=stack_name, TemplateBody=json.dumps(template))

    # then check list of things
    iot_conn = boto3.client("iot", region_name=TEST_REGION)
    assert len(iot_conn.list_job_templates()["jobTemplates"]) == 1

    # then check stack
    cfn_conn.delete_stack(StackName=stack_name)

    # and
    assert len(iot_conn.list_job_templates()["jobTemplates"]) == 0



from datetime import datetime

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


@mock_aws
def test_get_all_server_certs():
    conn = boto3.client("iam", region_name="us-east-1")

    conn.upload_server_certificate(
        ServerCertificateName="certname",
        CertificateBody="certbody",
        PrivateKey="privatekey",
    )
    certs = conn.list_server_certificates()["ServerCertificateMetadataList"]
    assert len(certs) == 1
    cert1 = certs[0]
    assert cert1["ServerCertificateName"] == "certname"
    assert cert1["Arn"] == f"arn:aws:iam::{ACCOUNT_ID}:server-certificate/certname"


@mock_aws
def test_get_server_cert_doesnt_exist():
    conn = boto3.client("iam", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        conn.get_server_certificate(ServerCertificateName="NonExistant")
    err = ex.value.response["Error"]
    assert err["Code"] == "NoSuchEntity"
    assert (
        err["Message"]
        == "The Server Certificate with name NonExistant cannot be found."
    )


@mock_aws
def test_get_server_cert():
    conn = boto3.client("iam", region_name="us-east-1")

    conn.upload_server_certificate(
        ServerCertificateName="certname",
        CertificateBody="certbody",
        PrivateKey="privatekey",
    )
    cert = conn.get_server_certificate(ServerCertificateName="certname")[
        "ServerCertificate"
    ]
    assert cert["CertificateBody"] == "certbody"
    assert "CertificateChain" not in cert
    assert "Tags" not in cert
    metadata = cert["ServerCertificateMetadata"]
    assert metadata["Path"] == "/"
    assert metadata["ServerCertificateName"] == "certname"
    assert metadata["Arn"] == f"arn:aws:iam::{ACCOUNT_ID}:server-certificate/certname"
    assert "ServerCertificateId" in metadata
    assert isinstance(metadata["UploadDate"], datetime)
    assert isinstance(metadata["Expiration"], datetime)


@mock_aws
def test_delete_server_cert():
    conn = boto3.client("iam", region_name="us-east-1")

    conn.upload_server_certificate(
        ServerCertificateName="certname",
        CertificateBody="certbody",
        PrivateKey="privatekey",
    )
    conn.get_server_certificate(ServerCertificateName="certname")
    conn.delete_server_certificate(ServerCertificateName="certname")

    with pytest.raises(ClientError) as ex:
        conn.get_server_certificate(ServerCertificateName="certname")
    err = ex.value.response["Error"]
    assert err["Code"] == "NoSuchEntity"
    assert (
        err["Message"] == "The Server Certificate with name certname cannot be found."
    )


@mock_aws
def test_delete_unknown_server_cert():
    conn = boto3.client("iam", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        conn.delete_server_certificate(ServerCertificateName="certname")
    err = ex.value.response["Error"]
    assert err["Code"] == "NoSuchEntity"
    assert (
        err["Message"] == "The Server Certificate with name certname cannot be found."
    )



import json

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse

from .models import GLOBAL_REGION, WAFV2Backend, wafv2_backends


class WAFV2Response(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="wafv2")

    @property
    def wafv2_backend(self) -> WAFV2Backend:
        return wafv2_backends[self.current_account][self.region]

    def associate_web_acl(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        web_acl_arn = body["WebACLArn"]
        resource_arn = body["ResourceArn"]
        self.wafv2_backend.associate_web_acl(web_acl_arn, resource_arn)
        return 200, {}, "{}"

    def disassociate_web_acl(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        resource_arn = body["ResourceArn"]
        self.wafv2_backend.disassociate_web_acl(resource_arn)
        return 200, {}, "{}"

    def get_web_acl_for_resource(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        resource_arn = body["ResourceArn"]
        web_acl = self.wafv2_backend.get_web_acl_for_resource(resource_arn)
        response = {"WebACL": web_acl.to_dict() if web_acl else None}
        response_headers = {"Content-Type": "application/json"}
        return 200, response_headers, json.dumps(response)

    def create_web_acl(self) -> TYPE_RESPONSE:
        """https://docs.aws.amazon.com/waf/latest/APIReference/API_CreateWebACL.html (response syntax section)"""

        scope = self._get_param("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        name = self._get_param("Name")
        body = json.loads(self.body)
        description = body.get("Description")
        tags = body.get("Tags", [])
        rules = body.get("Rules", [])
        association_config = body.get("AssociationConfig")
        captcha_config = body.get("CaptchaConfig")
        challenge_config = body.get("ChallengeConfig")
        custom_response_bodies = body.get("CustomResponseBodies")
        token_domains = body.get("TokenDomains")
        web_acl = self.wafv2_backend.create_web_acl(
            name,
            body["VisibilityConfig"],
            body["DefaultAction"],
            scope,
            description,
            tags,
            rules,
            association_config,
            captcha_config,
            challenge_config,
            custom_response_bodies,
            token_domains,
        )
        response = {"Summary": web_acl.to_short_dict()}
        response_headers = {"Content-Type": "application/json"}
        return 200, response_headers, json.dumps(response)

    def delete_web_acl(self) -> TYPE_RESPONSE:
        scope = self._get_param("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        name = self._get_param("Name")
        _id = self._get_param("Id")
        lock_token = self._get_param("LockToken")
        self.wafv2_backend.delete_web_acl(name, scope, _id, lock_token)
        response_headers = {"Content-Type": "application/json"}
        return 200, response_headers, "{}"

    def get_web_acl(self) -> TYPE_RESPONSE:
        scope = self._get_param("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        name = self._get_param("Name")
        _id = self._get_param("Id")
        web_acl = self.wafv2_backend.get_web_acl(name, _id)
        response = {"WebACL": web_acl.to_dict(), "LockToken": web_acl.lock_token}
        response_headers = {"Content-Type": "application/json"}
        return 200, response_headers, json.dumps(response)

    def list_web_ac_ls(self) -> TYPE_RESPONSE:
        """https://docs.aws.amazon.com/waf/latest/APIReference/API_ListWebACLs.html (response syntax section)"""

        scope = self._get_param("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        limit = self._get_int_param("Limit")
        next_marker = self._get_param("NextMarker")
        web_acls, next_marker = self.wafv2_backend.list_web_acls(
            limit=limit, next_marker=next_marker
        )
        response = {
            "NextMarker": next_marker,
            "WebACLs": [web.to_short_dict() for web in web_acls],
        }
        response_headers = {"Content-Type": "application/json"}
        return 200, response_headers, json.dumps(response)

    def list_rule_groups(self) -> TYPE_RESPONSE:
        scope = self._get_param("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        limit = self._get_int_param("Limit")
        next_marker = self._get_param("NextMarker")
        rule_groups, next_marker = self.wafv2_backend.list_rule_groups(
            scope, limit=limit, next_marker=next_marker
        )
        response = {
            "RuleGroups": [rg.to_short_dict() for rg in rule_groups],
            "NextMarker": next_marker,
        }
        response_headers = {"Content-Type": "application/json"}
        return 200, response_headers, json.dumps(response)

    def list_tags_for_resource(self) -> TYPE_RESPONSE:
        arn = self._get_param("ResourceARN")
        self.region = arn.split(":")[3]
        limit = self._get_int_param("Limit")
        next_marker = self._get_param("NextMarker")
        tags, next_marker = self.wafv2_backend.list_tags_for_resource(
            arn, limit=limit, next_marker=next_marker
        )
        response = {
            "TagInfoForResource": {"ResourceARN": arn, "TagList": tags},
            "NextMarker": next_marker,
        }
        response_headers = {"Content-Type": "application/json"}
        return 200, response_headers, json.dumps(response)

    def tag_resource(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        arn = body.get("ResourceARN")
        self.region = arn.split(":")[3]
        tags = body.get("Tags")
        self.wafv2_backend.tag_resource(arn, tags)
        return 200, {}, "{}"

    def untag_resource(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        arn = body.get("ResourceARN")
        self.region = arn.split(":")[3]
        tag_keys = body.get("TagKeys")
        self.wafv2_backend.untag_resource(arn, tag_keys)
        return 200, {}, "{}"

    def update_web_acl(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        name = body.get("Name")
        _id = body.get("Id")
        scope = body.get("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        default_action = body.get("DefaultAction")
        rules = body.get("Rules")
        description = body.get("Description")
        visibility_config = body.get("VisibilityConfig")
        lock_token = body.get("LockToken")
        custom_response_bodies = body.get("CustomResponseBodies")
        captcha_config = body.get("CaptchaConfig")
        challenge_config = body.get("ChallengeConfig")
        token_domains = body.get("TokenDomains")
        association_config = body.get("AssociationConfig")
        new_lock_token = self.wafv2_backend.update_web_acl(
            name,
            _id,
            default_action,
            rules,
            description,
            visibility_config,
            lock_token,
            custom_response_bodies,
            captcha_config,
            challenge_config,
            token_domains,
            association_config,
        )
        return 200, {}, json.dumps({"NextLockToken": new_lock_token})

    def create_ip_set(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)

        name = body.get("Name")
        scope = body.get("Scope")

        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION

        description = body.get("Description")
        ip_address_version = body.get("IPAddressVersion")
        addresses = body.get("Addresses")
        tags = body.get("Tags")

        ip_set = self.wafv2_backend.create_ip_set(
            name, scope, description, ip_address_version, addresses, tags
        )
        return (
            200,
            {},
            json.dumps(
                {
                    "Summary": {
                        "Name": ip_set.name,
                        "Id": ip_set.ip_set_id,
                        "Description": ip_set.description,
                        "LockToken": ip_set.lock_token,
                        "ARN": ip_set.arn,
                    }
                }
            ),
        )

    def delete_ip_set(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)

        name = body.get("Name")
        scope = body.get("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION

        _id = body.get("Id")
        lock_token = body.get("LockToken")

        self.wafv2_backend.delete_ip_set(name, scope, _id, lock_token)

        return 200, {}, "{}"

    def list_ip_sets(self) -> str:
        scope = self._get_param("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        next_marker = self._get_param("NextMarker")
        limit = self._get_int_param("Limit")
        ip_sets, next_token = self.wafv2_backend.list_ip_sets(
            scope, next_marker=next_marker, limit=limit
        )

        formatted_ip_sets = [
            {
                "Name": ip_set.name,
                "Id": ip_set.ip_set_id,
                "Description": ip_set.description,
                "LockToken": ip_set.lock_token,
                "ARN": ip_set.arn,
            }
            for ip_set in ip_sets
        ]

        return json.dumps({"NextMarker": next_token, "IPSets": formatted_ip_sets})

    def get_ip_set(self) -> TYPE_RESPONSE:
        scope = self._get_param("Scope")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION

        ip_set = self.wafv2_backend.get_ip_set(
            name=self._get_param("Name"), scope=scope, _id=self._get_param("Id")
        )

        dict_ip = ip_set.to_dict()
        lock_token = dict_ip.pop("LockToken")
        return 200, {}, json.dumps({"IPSet": dict_ip, "LockToken": lock_token})

    def update_ip_set(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)

        name = body.get("Name")
        scope = body.get("Scope")
        _id = body.get("Id")

        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION

        description = body.get("Description")
        addresses = body.get("Addresses")
        lock_token = body.get("LockToken")
        updated_ip_set = self.wafv2_backend.update_ip_set(
            name, scope, _id, description, addresses, lock_token
        )

        return 200, {}, json.dumps({"NextLockToken": updated_ip_set.lock_token})

    def put_logging_configuration(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        logging_configuration_parameter = body["LoggingConfiguration"]
        resource_arn = logging_configuration_parameter["ResourceArn"]
        log_destination_configs = logging_configuration_parameter[
            "LogDestinationConfigs"
        ]
        redacted_fields = logging_configuration_parameter.get("RedactedFields")
        managed_by_firewall_manager = logging_configuration_parameter.get(
            "ManagedByFirewallManager"
        )
        logging_filter = logging_configuration_parameter.get("LoggingFilter")
        logging_configuration = self.wafv2_backend.put_logging_configuration(
            resource_arn,
            log_destination_configs,
            redacted_fields,
            managed_by_firewall_manager,
            logging_filter,
        )
        return (
            200,
            {},
            json.dumps(
                {
                    "LoggingConfiguration": {
                        k: v
                        for k, v in logging_configuration.to_dict().items()
                        if v is not None
                    }
                }
            ),
        )

    def get_logging_configuration(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        resource_arn = body["ResourceArn"]
        logging_configuration = self.wafv2_backend.get_logging_configuration(
            resource_arn
        )
        return (
            200,
            {},
            json.dumps(
                {
                    "LoggingConfiguration": {
                        k: v
                        for k, v in logging_configuration.to_dict().items()
                        if v is not None
                    }
                }
            ),
        )

    def list_logging_configurations(self) -> str:
        body = json.loads(self.body)
        scope = body.get("Scope")
        limit = self._get_int_param("Limit")
        next_marker = self._get_param("NextMarker")
        log_configs, next_marker = self.wafv2_backend.list_logging_configurations(
            scope, limit=limit, next_marker=next_marker
        )

        formatted = [
            {k: v for k, v in config.to_dict().items() if v is not None}
            for config in log_configs
        ]
        return json.dumps(
            {"LoggingConfigurations": formatted, "NextMarker": next_marker}
        )

    def delete_logging_configuration(self) -> TYPE_RESPONSE:
        body = json.loads(self.body)
        resource_arn = body["ResourceArn"]
        self.wafv2_backend.delete_logging_configuration(resource_arn)
        return 200, {}, "{}"

    def create_rule_group(self) -> TYPE_RESPONSE:
        name = self._get_param("Name")
        scope = self._get_param("Scope")
        capacity = self._get_param("Capacity")
        description = self._get_param("Description")
        rules = self._get_param("Rules", [])
        visibility_config = self._get_param("VisibilityConfig")
        tags = self._get_param("Tags")
        custom_response_bodies = self._get_param("CustomResponseBodies")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        group = self.wafv2_backend.create_rule_group(
            name=name,
            scope=scope,
            capacity=capacity,
            description=description,
            rules=rules,
            visibility_config=visibility_config,
            tags=tags,
            custom_response_bodies=custom_response_bodies,
        )
        return 200, {}, json.dumps(dict(Summary=group.to_short_dict()))

    def update_rule_group(self) -> TYPE_RESPONSE:
        name = self._get_param("Name")
        scope = self._get_param("Scope")
        id = self._get_param("Id")
        description = self._get_param("Description")
        rules = self._get_param("Rules")
        visibility_config = self._get_param("VisibilityConfig")
        lock_token = self._get_param("LockToken")
        custom_response_bodies = self._get_param("CustomResponseBodies")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        updated_group = self.wafv2_backend.update_rule_group(
            name=name,
            scope=scope,
            id=id,
            description=description,
            rules=rules,
            visibility_config=visibility_config,
            lock_token=lock_token,
            custom_response_bodies=custom_response_bodies,
        )
        return 200, {}, json.dumps(dict(NextLockToken=updated_group.lock_token))

    def delete_rule_group(self) -> TYPE_RESPONSE:
        name = self._get_param("Name")
        scope = self._get_param("Scope")
        id = self._get_param("Id")
        lock_token = self._get_param("LockToken")
        if scope == "CLOUDFRONT":
            self.region = GLOBAL_REGION
        self.wafv2_backend.delete_rule_group(
            name=name,
            scope=scope,
            id=id,
            lock_token=lock_token,
        )
        return 200, {}, json.dumps(dict())

    def get_rule_group(self) -> TYPE_RESPONSE:
        name = self._get_param("Name")
        scope = self._get_param("Scope")
        id = self._get_param("Id")
        arn = self._get_param("ARN")
        if scope == "CLOUDFRONT" or (
            isinstance(arn, str) and arn.split(":")[3] == GLOBAL_REGION
        ):
            self.region = GLOBAL_REGION
        rule_group = self.wafv2_backend.get_rule_group(
            name=name,
            scope=scope,
            id=id,
            arn=arn,
        )
        group_dict = rule_group.to_dict()
        lock_token = group_dict.pop("LockToken")
        return 200, {}, json.dumps(dict(RuleGroup=group_dict, LockToken=lock_token))


# notes about region and scope
# --scope = CLOUDFRONT is ALWAYS us-east-1 (but we use "global" instead to differentiate between REGIONAL us-east-1)
# --scope = REGIONAL defaults to us-east-1, but could be anything if specified with --region=<anyRegion>
# region is grabbed from the auth header, NOT from the body - even with --region flag
# The CLOUDFRONT wacls in aws console are located in us-east-1 but the us-east-1 REGIONAL wacls are not included



import datetime
from typing import Any, Dict, List, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import unix_time
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition

from .data import compatible_versions
from .exceptions import EngineTypeNotFoundException, ResourceNotFoundException

default_cluster_config = {
    "InstanceType": "t3.small.search",
    "InstanceCount": 1,
    "DedicatedMasterEnabled": False,
    "ZoneAwarenessEnabled": False,
    "WarmEnabled": False,
    "ColdStorageOptions": {"Enabled": False},
}
default_advanced_security_options = {
    "Enabled": False,
    "InternalUserDatabaseEnabled": False,
    "AnonymousAuthEnabled": False,
}
default_domain_endpoint_options = {
    "EnforceHTTPS": False,
    "TLSSecurityPolicy": "Policy-Min-TLS-1-0-2019-07",
    "CustomEndpointEnabled": False,
}
default_software_update_options = {
    "CurrentVersion": "",
    "NewVersion": "",
    "UpdateAvailable": False,
    "Cancellable": False,
    "UpdateStatus": "COMPLETED",
    "Description": "There is no software update available for this domain.",
    "AutomatedUpdateDate": "1969-12-31T23:00:00-01:00",
    "OptionalDeployment": True,
}
default_advanced_options = {
    "override_main_response_version": "false",
    "rest.action.multi.allow_explicit_index": "true",
}


class EngineVersion(BaseModel):
    def __init__(self, options: str, create_time: datetime.datetime) -> None:
        self.options = options or "OpenSearch_2.5"
        self.create_time = unix_time(create_time)
        self.update_time = self.create_time

    def to_dict(self) -> Dict[str, Any]:
        return {
            "Options": self.options,
            "Status": {
                "CreationDate": self.create_time,
                "PendingDeletion": False,
                "State": "Active",
                "UpdateDate": self.update_time,
                "UpdateVersion": 28,
            },
        }


class OpenSearchDomain(BaseModel):
    def __init__(
        self,
        account_id: str,
        region: str,
        domain_name: str,
        engine_version: str,
        cluster_config: Dict[str, Any],
        ebs_options: Dict[str, Any],
        access_policies: str,
        snapshot_options: Dict[str, int],
        vpc_options: Dict[str, List[str]],
        cognito_options: Dict[str, Any],
        encryption_at_rest_options: Dict[str, Any],
        node_to_node_encryption_options: Dict[str, bool],
        advanced_options: Dict[str, str],
        log_publishing_options: Dict[str, Any],
        domain_endpoint_options: Dict[str, Any],
        advanced_security_options: Dict[str, Any],
        auto_tune_options: Dict[str, Any],
        off_peak_window_options: Dict[str, Any],
        software_update_options: Dict[str, bool],
        is_es: bool,
        elasticsearch_version: Optional[str],
        elasticsearch_cluster_config: Optional[str],
    ):
        self.domain_id = f"{account_id}/{domain_name}"
        self.domain_name = domain_name
        self.arn = (
            f"arn:{get_partition(region)}:es:{region}:{account_id}:domain/{domain_name}"
        )
        self.engine_version = EngineVersion(engine_version, datetime.datetime.now())
        self.cluster_config = cluster_config or {}
        self.ebs_options = ebs_options or {"EBSEnabled": False}
        self.access_policies = access_policies or ""
        self.snapshot_options = snapshot_options or {"AutomatedSnapshotStartHour": 0}
        self.vpc_options = vpc_options
        self.cognito_options = cognito_options or {"Enabled": False}
        self.encryption_at_rest_options = encryption_at_rest_options or {
            "Enabled": False
        }
        self.node_to_node_encryption_options = node_to_node_encryption_options or {
            "Enabled": False
        }
        self.advanced_options = advanced_options or default_advanced_options
        self.log_publishing_options = log_publishing_options
        self.domain_endpoint_options = (
            domain_endpoint_options or default_domain_endpoint_options
        )
        self.advanced_security_options = (
            advanced_security_options or default_advanced_security_options
        )
        self.auto_tune_options = auto_tune_options or {"State": "ENABLE_IN_PROGRESS"}
        if not self.auto_tune_options.get("State"):
            self.auto_tune_options["State"] = "ENABLED"
        self.off_peak_windows_options = off_peak_window_options
        self.software_update_options = (
            software_update_options or default_software_update_options
        )
        self.engine_type = "Elasticsearch" if is_es else "OpenSearch"
        self.is_es = is_es
        self.elasticsearch_version = elasticsearch_version
        self.elasticsearch_cluster_config = elasticsearch_cluster_config

        self.deleted = False
        self.processing = False

        # Defaults
        for key, value in default_cluster_config.items():
            if key not in self.cluster_config:
                self.cluster_config[key] = value

        if self.vpc_options is None:
            self.endpoint: Optional[str] = f"{domain_name}.{region}.es.amazonaws.com"
            self.endpoints: Optional[Dict[str, str]] = None
        else:
            self.endpoint = None
            self.endpoints = {"vpc": f"{domain_name}.{region}.es.amazonaws.com"}

    def delete(self) -> None:
        self.deleted = True
        self.processing = True

    def dct_options(self) -> Dict[str, Any]:
        return {
            "Endpoint": self.endpoint,
            "Endpoints": self.endpoints,
            "EngineVersion": self.engine_version.to_dict(),
            "ClusterConfig": self.cluster_config,
            "EBSOptions": self.ebs_options,
            "AccessPolicies": self.access_policies,
            "SnapshotOptions": self.snapshot_options,
            "VPCOptions": self.vpc_options,
            "CognitoOptions": self.cognito_options,
            "EncryptionAtRestOptions": self.encryption_at_rest_options,
            "NodeToNodeEncryptionOptions": self.node_to_node_encryption_options,
            "AdvancedOptions": self.advanced_options,
            "LogPublishingOptions": self.log_publishing_options,
            "DomainEndpointOptions": self.domain_endpoint_options,
            "AdvancedSecurityOptions": self.advanced_security_options,
            "AutoTuneOptions": self.auto_tune_options,
            "OffPeakWindowsOptions": self.off_peak_windows_options,
            "SoftwareUpdateOptions": self.software_update_options,
            "ElasticsearchVersion": self.elasticsearch_version,
            "ElasticsearchClusterConfig": self.elasticsearch_cluster_config,
        }

    def to_dict(self) -> Dict[str, Any]:
        dct = {
            "DomainId": self.domain_id,
            "DomainName": self.domain_name,
            "ARN": self.arn,
            "Created": True,
            "Deleted": self.deleted,
            "Processing": self.processing,
            "UpgradeProcessing": False,
        }
        for key, value in self.dct_options().items():
            if value is not None:
                dct[key] = value
        return dct

    def to_config_dict(self) -> Dict[str, Any]:
        dct: Dict[str, Any] = dict()
        for key, value in self.dct_options().items():
            if value is not None:
                dct[key] = {"Options": value}
        return dct

    def update(
        self,
        cluster_config: Dict[str, Any],
        ebs_options: Dict[str, Any],
        access_policies: str,
        snapshot_options: Dict[str, int],
        vpc_options: Dict[str, List[str]],
        cognito_options: Dict[str, Any],
        encryption_at_rest_options: Dict[str, Any],
        node_to_node_encryption_options: Dict[str, bool],
        advanced_options: Dict[str, str],
        log_publishing_options: Dict[str, Any],
        domain_endpoint_options: Dict[str, Any],
        advanced_security_options: Dict[str, Any],
        auto_tune_options: Dict[str, Any],
        off_peak_window_options: Dict[str, Any],
        software_update_options: Dict[str, bool],
    ) -> None:
        self.cluster_config = cluster_config or self.cluster_config
        self.ebs_options = ebs_options or self.ebs_options
        self.access_policies = access_policies or self.access_policies
        self.snapshot_options = snapshot_options or self.snapshot_options
        self.vpc_options = vpc_options or self.vpc_options
        self.cognito_options = cognito_options or self.cognito_options
        self.encryption_at_rest_options = (
            encryption_at_rest_options or self.encryption_at_rest_options
        )
        self.node_to_node_encryption_options = (
            node_to_node_encryption_options or self.node_to_node_encryption_options
        )
        self.advanced_options = advanced_options or self.advanced_options
        self.log_publishing_options = (
            log_publishing_options or self.log_publishing_options
        )
        self.domain_endpoint_options = (
            domain_endpoint_options or self.domain_endpoint_options
        )
        self.advanced_security_options = (
            advanced_security_options or self.advanced_security_options
        )
        self.auto_tune_options = auto_tune_options or self.auto_tune_options
        self.off_peak_windows_options = (
            off_peak_window_options or self.off_peak_windows_options
        )
        self.software_update_options = (
            software_update_options or self.software_update_options
        )
        self.engine_version.update_time = unix_time(datetime.datetime.now())


class OpenSearchServiceBackend(BaseBackend):
    """Implementation of OpenSearchService APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.domains: Dict[str, OpenSearchDomain] = dict()
        self.tagger = TaggingService()

    def create_domain(
        self,
        domain_name: str,
        engine_version: str,
        cluster_config: Dict[str, Any],
        ebs_options: Dict[str, Any],
        access_policies: str,
        snapshot_options: Dict[str, Any],
        vpc_options: Dict[str, Any],
        cognito_options: Dict[str, Any],
        encryption_at_rest_options: Dict[str, Any],
        node_to_node_encryption_options: Dict[str, Any],
        advanced_options: Dict[str, Any],
        log_publishing_options: Dict[str, Any],
        domain_endpoint_options: Dict[str, Any],
        advanced_security_options: Dict[str, Any],
        tag_list: List[Dict[str, str]],
        auto_tune_options: Dict[str, Any],
        off_peak_window_options: Dict[str, Any],
        software_update_options: Dict[str, Any],
        is_es: bool,
        elasticsearch_version: Optional[str],
        elasticsearch_cluster_config: Optional[str],
    ) -> OpenSearchDomain:
        domain = OpenSearchDomain(
            account_id=self.account_id,
            region=self.region_name,
            domain_name=domain_name,
            engine_version=engine_version,
            cluster_config=cluster_config,
            ebs_options=ebs_options,
            access_policies=access_policies,
            snapshot_options=snapshot_options,
            vpc_options=vpc_options,
            cognito_options=cognito_options,
            encryption_at_rest_options=encryption_at_rest_options,
            node_to_node_encryption_options=node_to_node_encryption_options,
            advanced_options=advanced_options,
            log_publishing_options=log_publishing_options,
            domain_endpoint_options=domain_endpoint_options,
            advanced_security_options=advanced_security_options,
            auto_tune_options=auto_tune_options,
            off_peak_window_options=off_peak_window_options,
            software_update_options=software_update_options,
            is_es=is_es,
            elasticsearch_version=elasticsearch_version,
            elasticsearch_cluster_config=elasticsearch_cluster_config,
        )
        self.domains[domain_name] = domain
        if tag_list:
            self.add_tags(domain.arn, tag_list)
        return domain

    def get_compatible_versions(self, domain_name: str) -> List[Dict[str, Any]]:
        if domain_name not in self.domains:
            raise ResourceNotFoundException(domain_name)
        return compatible_versions

    def delete_domain(self, domain_name: str) -> OpenSearchDomain:
        if domain_name not in self.domains:
            raise ResourceNotFoundException(domain_name)
        self.domains[domain_name].delete()
        return self.domains.pop(domain_name)

    def describe_domain(self, domain_name: str) -> OpenSearchDomain:
        if domain_name not in self.domains:
            raise ResourceNotFoundException(domain_name)
        return self.domains[domain_name]

    def describe_domain_config(self, domain_name: str) -> OpenSearchDomain:
        return self.describe_domain(domain_name)

    def update_domain_config(
        self,
        domain_name: str,
        cluster_config: Dict[str, Any],
        ebs_options: Dict[str, Any],
        access_policies: str,
        snapshot_options: Dict[str, Any],
        vpc_options: Dict[str, Any],
        cognito_options: Dict[str, Any],
        encryption_at_rest_options: Dict[str, Any],
        node_to_node_encryption_options: Dict[str, Any],
        advanced_options: Dict[str, Any],
        log_publishing_options: Dict[str, Any],
        domain_endpoint_options: Dict[str, Any],
        advanced_security_options: Dict[str, Any],
        auto_tune_options: Dict[str, Any],
        off_peak_window_options: Dict[str, Any],
        software_update_options: Dict[str, Any],
    ) -> OpenSearchDomain:
        domain = self.describe_domain(domain_name)
        domain.update(
            cluster_config=cluster_config,
            ebs_options=ebs_options,
            access_policies=access_policies,
            snapshot_options=snapshot_options,
            vpc_options=vpc_options,
            cognito_options=cognito_options,
            encryption_at_rest_options=encryption_at_rest_options,
            node_to_node_encryption_options=node_to_node_encryption_options,
            advanced_options=advanced_options,
            log_publishing_options=log_publishing_options,
            domain_endpoint_options=domain_endpoint_options,
            advanced_security_options=advanced_security_options,
            auto_tune_options=auto_tune_options,
            off_peak_window_options=off_peak_window_options,
            software_update_options=software_update_options,
        )
        return domain

    def add_tags(self, arn: str, tags: List[Dict[str, str]]) -> None:
        self.tagger.tag_resource(arn, tags)

    def list_tags(self, arn: str) -> List[Dict[str, str]]:
        return self.tagger.list_tags_for_resource(arn)["Tags"]

    def remove_tags(self, arn: str, tag_keys: List[str]) -> None:
        self.tagger.untag_resource_using_names(arn, tag_keys)

    def list_domain_names(self, engine_type: str) -> List[Dict[str, str]]:
        domains = []
        if engine_type and engine_type not in ["Elasticsearch", "OpenSearch"]:
            raise EngineTypeNotFoundException(engine_type)
        for domain in self.domains.values():
            if engine_type:
                if engine_type == domain.engine_type:
                    domains.append(
                        {"DomainName": domain.domain_name, "EngineType": engine_type}
                    )
            else:
                domains.append(
                    {"DomainName": domain.domain_name, "EngineType": domain.engine_type}
                )
        return domains

    def describe_domains(self, domain_names: List[str]) -> List[OpenSearchDomain]:
        queried_domains = []
        for domain_name in domain_names:
            if domain_name in self.domains:
                queried_domains.append(self.domains[domain_name])
        return queried_domains


opensearch_backends = BackendDict(OpenSearchServiceBackend, "opensearch")



from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional

from moto.appmesh.dataclasses.shared import (
    Duration,
    Metadata,
    MissingField,
    Status,
    Timeout,
)
from moto.appmesh.utils.common import clean_dict


@dataclass
class CertificateFile:
    certificate_chain: str

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {"certificateChain": self.certificate_chain}


@dataclass
class CertificateFileWithPrivateKey(CertificateFile):
    private_key: str

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {
            "certificateChain": self.certificate_chain,
            "privateKey": self.private_key,
        }


@dataclass
class SDS:
    secret_name: str

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {"secretName": self.secret_name}


@dataclass
class Certificate:
    file: Optional[CertificateFileWithPrivateKey]
    sds: Optional[SDS]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "file": (self.file or MissingField()).to_dict(),
                "sds": (self.sds or MissingField()).to_dict(),
            }
        )


@dataclass
class ListenerCertificateACM:
    certificate_arn: str

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {"certificateArn": self.certificate_arn}


@dataclass
class TLSListenerCertificate(Certificate):
    acm: Optional[ListenerCertificateACM]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "acm": (self.acm or MissingField()).to_dict(),
                "file": (self.file or MissingField()).to_dict(),
                "sds": (self.sds or MissingField()).to_dict(),
            }
        )


@dataclass
class Match:
    exact: List[str]

    to_dict = asdict


@dataclass
class SubjectAlternativeNames:
    match: Match

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {"match": self.match.to_dict()}


@dataclass
class ACM:
    certificate_authority_arns: List[str]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {"certificateAuthorityArns": self.certificate_authority_arns}


@dataclass
class Trust:
    file: Optional[CertificateFile]
    sds: Optional[SDS]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "file": (self.file or MissingField()).to_dict(),
                "sds": (self.sds or MissingField()).to_dict(),
            }
        )


@dataclass
class BackendTrust(Trust):
    acm: Optional[ACM]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "acm": (self.acm or MissingField()).to_dict(),
                "file": (self.file or MissingField()).to_dict(),
                "sds": (self.sds or MissingField()).to_dict(),
            }
        )


@dataclass
class Validation:
    subject_alternative_names: Optional[SubjectAlternativeNames]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "subjectAlternativeNames": (
                    self.subject_alternative_names or MissingField()
                ).to_dict()
            }
        )


@dataclass
class TLSBackendValidation(Validation):
    trust: BackendTrust

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "subjectAlternativeNames": (
                    self.subject_alternative_names or MissingField()
                ).to_dict(),
                "trust": self.trust.to_dict(),
            }
        )


@dataclass
class TLSListenerValidation(Validation):
    trust: Trust

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "subjectAlternativeNames": (
                    self.subject_alternative_names or MissingField()
                ).to_dict(),
                "trust": self.trust.to_dict(),
            }
        )


@dataclass
class TLSClientPolicy:
    certificate: Optional[Certificate]
    enforce: Optional[bool]
    ports: Optional[List[int]]
    validation: TLSBackendValidation

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "certificate": (self.certificate or MissingField()).to_dict(),
                "enforce": self.enforce,
                "ports": self.ports,
                "validation": self.validation.to_dict(),
            }
        )


@dataclass
class ClientPolicy:
    tls: Optional[TLSClientPolicy]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict({"tls": (self.tls or MissingField()).to_dict()})


@dataclass
class BackendDefaults:
    client_policy: Optional[ClientPolicy]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {"clientPolicy": (self.client_policy or MissingField()).to_dict()}
        )


@dataclass
class VirtualService:
    client_policy: Optional[ClientPolicy]
    virtual_service_name: str

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "clientPolicy": (self.client_policy or MissingField()).to_dict(),
                "virtualServiceName": self.virtual_service_name,
            }
        )


@dataclass
class Backend:
    virtual_service: Optional[VirtualService]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {"virtualService": (self.virtual_service or MissingField()).to_dict()}
        )


@dataclass
class HTTPConnection:
    max_connections: int
    max_pending_requests: Optional[int]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "maxConnections": self.max_connections,
                "maxPendingRequests": self.max_pending_requests,
            }
        )


@dataclass
class GRPCOrHTTP2Connection:
    max_requests: int

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {"maxRequests": self.max_requests}


@dataclass
class TCPConnection:
    max_connections: int

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {"maxConnections": self.max_connections}


@dataclass
class ConnectionPool:
    grpc: Optional[GRPCOrHTTP2Connection]
    http: Optional[HTTPConnection]
    http2: Optional[GRPCOrHTTP2Connection]
    tcp: Optional[TCPConnection]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "grpc": (self.grpc or MissingField()).to_dict(),
                "http": (self.http or MissingField()).to_dict(),
                "http2": (self.http2 or MissingField()).to_dict(),
                "tcp": (self.tcp or MissingField()).to_dict(),
            }
        )


@dataclass
class HealthCheck:
    healthy_threshold: int
    interval_millis: int
    path: Optional[str]
    port: Optional[int]
    protocol: str
    timeout_millis: int
    unhealthy_threshold: int

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "healthyThreshold": self.healthy_threshold,
                "intervalMillis": self.interval_millis,
                "path": self.path,
                "port": self.port,
                "protocol": self.protocol,
                "timeoutMillis": self.timeout_millis,
                "unhealthyThreshold": self.unhealthy_threshold,
            }
        )


@dataclass
class OutlierDetection:
    base_ejection_duration: Duration
    interval: Duration
    max_ejection_percent: int
    max_server_errors: int

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {
            "baseEjectionDuration": self.base_ejection_duration.to_dict(),
            "interval": self.interval.to_dict(),
            "maxEjectionPercent": self.max_ejection_percent,
            "maxServerErrors": self.max_server_errors,
        }


@dataclass
class PortMapping:
    port: int
    protocol: str
    to_dict = asdict


@dataclass
class TCPTimeout:
    idle: Duration

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return {"idle": self.idle.to_dict()}


@dataclass
class ProtocolTimeouts:
    grpc: Optional[Timeout]
    http: Optional[Timeout]
    http2: Optional[Timeout]
    tcp: Optional[TCPTimeout]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "grpc": (self.grpc or MissingField()).to_dict(),
                "http": (self.http or MissingField()).to_dict(),
                "http2": (self.http2 or MissingField()).to_dict(),
                "tcp": (self.tcp or MissingField()).to_dict(),
            }
        )


@dataclass
class ListenerTLS:
    certificate: TLSListenerCertificate
    mode: str
    validation: Optional[TLSListenerValidation]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "certificate": self.certificate.to_dict(),
                "mode": self.mode,
                "validation": (self.validation or MissingField()).to_dict(),
            }
        )


@dataclass
class Listener:
    connection_pool: Optional[ConnectionPool]
    health_check: Optional[HealthCheck]
    outlier_detection: Optional[OutlierDetection]
    port_mapping: PortMapping
    timeout: Optional[ProtocolTimeouts]
    tls: Optional[ListenerTLS]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "connectionPool": (self.connection_pool or MissingField()).to_dict(),
                "healthCheck": (self.health_check or MissingField()).to_dict(),
                "outlierDetection": (
                    self.outlier_detection or MissingField()
                ).to_dict(),
                "portMapping": self.port_mapping.to_dict(),
                "timeout": (self.timeout or MissingField()).to_dict(),
                "tls": (self.tls or MissingField()).to_dict(),
            }
        )


@dataclass
class KeyValue:
    key: str
    value: str
    to_dict = asdict


@dataclass
class LoggingFormat:
    json: Optional[List[KeyValue]]
    text: Optional[str]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {"json": [pair.to_dict() for pair in self.json or []], "text": self.text}
        )


@dataclass
class AccessLogFile:
    format: Optional[LoggingFormat]
    path: str

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {"format": (self.format or MissingField()).to_dict(), "path": self.path}
        )


@dataclass
class AccessLog:
    file: Optional[AccessLogFile]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict({"file": (self.file or MissingField()).to_dict()})


@dataclass
class Logging:
    access_log: Optional[AccessLog]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict({"accessLog": (self.access_log or MissingField()).to_dict()})


@dataclass
class AWSCloudMap:
    attributes: Optional[List[KeyValue]]
    ip_preference: Optional[str]
    namespace_name: str
    service_name: str

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "attributes": [
                    attribute.to_dict() for attribute in self.attributes or []
                ],
                "ipPreference": self.ip_preference,
                "namespaceName": self.namespace_name,
                "serviceName": self.service_name,
            }
        )


@dataclass
class DNS:
    hostname: str
    ip_preference: Optional[str]
    response_type: Optional[str]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "hostname": self.hostname,
                "ipPreference": self.ip_preference,
                "responseType": self.response_type,
            }
        )


@dataclass
class ServiceDiscovery:
    aws_cloud_map: Optional[AWSCloudMap]
    dns: Optional[DNS]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "awsCloudMap": (self.aws_cloud_map or MissingField()).to_dict(),
                "dns": (self.dns or MissingField()).to_dict(),
            }
        )


@dataclass
class VirtualNodeSpec:
    backend_defaults: Optional[BackendDefaults]
    backends: Optional[List[Backend]]
    listeners: Optional[List[Listener]]
    logging: Optional[Logging]
    service_discovery: Optional[ServiceDiscovery]

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "backendDefaults": (self.backend_defaults or MissingField()).to_dict(),
                "backends": [backend.to_dict() for backend in self.backends or []],
                "listeners": [listener.to_dict() for listener in self.listeners or []],
                "logging": (self.logging or MissingField()).to_dict(),
                "serviceDiscovery": (
                    self.service_discovery or MissingField()
                ).to_dict(),
            }
        )


@dataclass
class VirtualNodeMetadata(Metadata):
    mesh_name: str = field(default="")
    virtual_node_name: str = field(default="")

    def __post_init__(self) -> None:
        if self.mesh_name == "":
            raise TypeError("__init__ missing 1 required argument: 'mesh_name'")
        if self.mesh_owner == "":
            raise TypeError("__init__ missing 1 required argument: 'route_name'")
        if self.virtual_node_name == "":
            raise TypeError("__init__ missing 1 required argument: 'virtual_node_name'")

    def formatted_for_list_api(self) -> Dict[str, Any]:  # type: ignore
        return {
            "arn": self.arn,
            "createdAt": self.created_at.strftime("%d/%m/%Y, %H:%M:%S"),
            "lastUpdatedAt": self.last_updated_at.strftime("%d/%m/%Y, %H:%M:%S"),
            "meshName": self.mesh_name,
            "meshOwner": self.mesh_owner,
            "resourceOwner": self.resource_owner,
            "version": self.version,
            "virtualNodeName": self.virtual_node_name,
        }

    def formatted_for_crud_apis(self) -> Dict[str, Any]:  # type: ignore
        return {
            "arn": self.arn,
            "createdAt": self.created_at.strftime("%d/%m/%Y, %H:%M:%S"),
            "lastUpdatedAt": self.last_updated_at.strftime("%d/%m/%Y, %H:%M:%S"),
            "meshOwner": self.mesh_owner,
            "resourceOwner": self.resource_owner,
            "uid": self.uid,
            "version": self.version,
        }


@dataclass
class VirtualNode:
    mesh_name: str
    mesh_owner: str
    metadata: VirtualNodeMetadata
    spec: VirtualNodeSpec
    virtual_node_name: str
    status: Status = field(default_factory=lambda: {"status": "ACTIVE"})
    tags: List[Dict[str, str]] = field(default_factory=list)

    def to_dict(self) -> Dict[str, Any]:  # type: ignore[misc]
        return clean_dict(
            {
                "meshName": self.mesh_name,
                "metadata": self.metadata.formatted_for_crud_apis(),
                "spec": self.spec.to_dict(),
                "status": self.status,
                "virtualNodeName": self.virtual_node_name,
            }
        )



import unittest

import boto3
import requests

import moto


# https://github.com/getmoto/moto/issues/6592
class TestDifferentAccountsDoesNotBreakSeeding:
    def setup_method(self) -> None:
        if not moto.settings.TEST_DECORATOR_MODE:
            raise unittest.SkipTest(
                "Seed behaviour only needs to be tested in DecoratorMode"
            )
        self.mock = moto.mock_aws()
        self.mock.start()

        requests.post("http://motoapi.amazonaws.com/moto-api/seed?a=42")

        self.sts_client = boto3.client("sts", "us-east-1")
        self.ec2_client = boto3.client("ec2", "us-east-1")

        self.session_client = self.sts_client.assume_role(
            RoleArn="arn:aws:iam::111111111111:role/my-role",
            RoleSessionName="role-session-name",
        )

        self.session = boto3.session.Session(
            aws_access_key_id=self.session_client["Credentials"]["AccessKeyId"],
            aws_secret_access_key=self.session_client["Credentials"]["SecretAccessKey"],
            aws_session_token=self.session_client["Credentials"]["SessionToken"],
        )

    def teardown_method(self) -> None:
        self.mock.stop()

    def test_0(self) -> None:
        # We seeded Moto in the setup-method
        # So our instances should have a fixed InstanceId
        instances = self.ec2_client.run_instances(MaxCount=1, MinCount=1)["Instances"]

        instance_ids = [instance["InstanceId"] for instance in instances]
        assert instance_ids == ["i-6e8b5359309cc6273"]

    def test_1(self) -> None:
        # Create some data in a different account (111111111111)
        self.sts_client.assume_role(
            RoleArn="arn:aws:iam::111111111111:role/my-role",
            RoleSessionName="role-session-name",
        )

    def test_2(self) -> None:
        # The fact that some data now exists in a different account (111111111111)
        # should not change the fixed InstanceId-value
        self.test_0()



from typing import Any, Optional

from moto.stepfunctions.parser.api import Arn, Definition, LongArn, StateMachineType


class AWSExecutionDetails:
    account: str
    region: str
    role_arn: str

    def __init__(self, account: str, region: str, role_arn: str):
        self.account = account
        self.region = region
        self.role_arn = role_arn


class ExecutionDetails:
    arn: LongArn
    name: str
    role_arn: Arn
    inpt: Optional[Any]
    start_time: str

    def __init__(
        self,
        arn: LongArn,
        name: str,
        role_arn: Arn,
        inpt: Optional[Any],
        start_time: str,
    ):
        self.arn = arn
        self.name = name
        self.role_arn = role_arn
        self.inpt = inpt
        self.start_time = start_time


class StateMachineDetails:
    arn: Arn
    name: str
    typ: StateMachineType
    definition: Definition

    def __init__(self, arn: Arn, name: str, typ: StateMachineType, definition: str):
        self.arn = arn
        self.name = name
        self.typ = typ
        self.definition = definition


class EvaluationDetails:
    aws_execution_details: AWSExecutionDetails
    execution_details: ExecutionDetails
    state_machine_details: StateMachineDetails

    def __init__(
        self,
        aws_execution_details: AWSExecutionDetails,
        execution_details: ExecutionDetails,
        state_machine_details: StateMachineDetails,
    ):
        self.aws_execution_details = aws_execution_details
        self.execution_details = execution_details
        self.state_machine_details = state_machine_details



"""Unit tests for dsql-supported APIs."""

from datetime import datetime

import boto3
from botocore.exceptions import ClientError
from dateutil.tz import tzutc
from freezegun import freeze_time

from moto import mock_aws, settings

TEST_REGION = "us-east-1"


@mock_aws
def test_create_cluster():
    client = boto3.client("dsql", region_name=TEST_REGION)
    with freeze_time("2024-12-22 12:34:00"):
        resp = client.create_cluster()

    identifier = resp["identifier"]
    assert identifier is not None
    assert resp["arn"] == f"arn:aws:dsql:us-east-1:123456789012:cluster/{identifier}"
    assert resp["deletionProtectionEnabled"] is True
    assert resp["status"] == "CREATING"
    if not settings.TEST_SERVER_MODE:
        assert resp["creationTime"] == datetime(2024, 12, 22, 12, 34, tzinfo=tzutc())


@mock_aws
def test_get_invalid_cluster():
    client = boto3.client("dsql", region_name=TEST_REGION)

    try:
        client.get_cluster(identifier="invalid")
    except ClientError as err:
        assert err.response["Error"]["Code"] == "ValidationException"
        assert err.response["Error"]["Message"] == "invalid Cluster Id"


@mock_aws
def test_get_cluster():
    client = boto3.client("dsql", region_name=TEST_REGION)
    with freeze_time("2024-12-22 12:34:00"):
        resp = client.create_cluster()

    identifier = resp["identifier"]

    get_resp = client.get_cluster(identifier=identifier)

    # TODO Add `witnessRegion` and `linkedClusterArns` when implement create-multi-region-clusters
    assert get_resp["identifier"] == identifier
    assert (
        get_resp["arn"] == f"arn:aws:dsql:us-east-1:123456789012:cluster/{identifier}"
    )
    assert get_resp["deletionProtectionEnabled"] is True
    assert get_resp["status"] == "CREATING"
    if not settings.TEST_SERVER_MODE:
        assert get_resp["creationTime"] == datetime(
            2024, 12, 22, 12, 34, tzinfo=tzutc()
        )


@mock_aws()
def test_generate_tokens():
    client = boto3.client("dsql", TEST_REGION)

    hostname = "dsql.amazonaws.com"

    admin_url = client.generate_db_connect_admin_auth_token(Hostname=hostname)
    assert admin_url.startswith(hostname)
    assert "Action=DbConnectAdmin" in admin_url

    url = client.generate_db_connect_auth_token(Hostname=hostname)
    assert url.startswith(hostname)
    assert "Action=DbConnect" in url



from typing import Optional

from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.parargs import Parameters
from moto.stepfunctions.parser.asl.component.common.timeouts.timeout import (
    EvalTimeoutError,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_selector import (
    ItemSelector,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.itemprocessor.map_run_record import (
    MapRunRecord,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.iterator.inline_iterator_worker import (
    InlineIteratorWorker,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_map.iteration.job import (
    Job,
    JobPool,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.program_state import (
    ProgramError,
    ProgramState,
    ProgramStopped,
)
from moto.stepfunctions.parser.asl.eval.states import ItemData, MapData


class DistributedIteratorWorker(InlineIteratorWorker):
    _map_run_record: MapRunRecord

    def __init__(
        self,
        work_name: str,
        job_pool: JobPool,
        env: Environment,
        parameters: Optional[Parameters],
        map_run_record: MapRunRecord,
        item_selector: Optional[ItemSelector],
    ):
        super().__init__(
            work_name=work_name,
            job_pool=job_pool,
            env=env,
            parameters=parameters,
            item_selector=item_selector,
        )
        self._map_run_record = map_run_record

    def _eval_job(self, env: Environment, job: Job) -> None:
        self._map_run_record.item_counter.total.count()
        self._map_run_record.item_counter.running.count()

        self._map_run_record.execution_counter.total.count()
        self._map_run_record.execution_counter.running.count()

        job_output = None
        try:
            env.states.context_object.context_object_data["Map"] = MapData(
                Item=ItemData(Index=job.job_index, Value=job.job_input)
            )

            env.states.reset(input_value=job.job_input)
            env.stack.append(env.states.get_input())
            self._eval_input(env_frame=env)

            job.job_program.eval(env)

            # TODO: verify behaviour with all of these scenarios.
            end_program_state: ProgramState = env.program_state()
            if isinstance(end_program_state, ProgramError):
                self._map_run_record.execution_counter.failed.count()
                self._map_run_record.item_counter.failed.count()
                job_output = None
            elif isinstance(end_program_state, ProgramStopped):
                self._map_run_record.execution_counter.aborted.count()
                self._map_run_record.item_counter.aborted.count()
            else:
                self._map_run_record.item_counter.succeeded.count()
                self._map_run_record.item_counter.results_written.count()

                self._map_run_record.execution_counter.succeeded.count()
                self._map_run_record.execution_counter.results_written.count()
                self._map_run_record.execution_counter.running.offset(-1)

                job_output = env.states.get_input()

        except EvalTimeoutError:
            self._map_run_record.item_counter.timed_out.count()

        except FailureEventException:
            self._map_run_record.item_counter.failed.count()

        except Exception:
            self._map_run_record.item_counter.failed.count()

        finally:
            self._map_run_record.item_counter.running.offset(-1)
            job.job_output = job_output

    def _eval_pool(self, job: Optional[Job], worker_frame: Environment) -> None:
        # Note: the frame has to be closed before the job, to ensure the owner environment is correctly updated
        #  before the evaluation continues; map states await for job termination not workers termination.
        if job is None:
            self._env.delete_frame(worker_frame)
            return

        # Evaluate the job.
        job_frame = worker_frame.open_frame()
        self._eval_job(env=job_frame, job=job)
        worker_frame.delete_frame(job_frame)

        # Evaluation terminates here due to exception in job, or worker was stopped.
        if isinstance(job.job_output, Exception) or self.stopped():
            self._env.delete_frame(worker_frame)
            self._job_pool.close_job(job)
            return

        next_job: Job = self._job_pool.next_job()
        # Iteration will terminate after this job.
        if next_job is None:
            self._env.delete_frame(worker_frame)
            self._job_pool.close_job(job)
            return

        self._job_pool.close_job(job)
        self._eval_pool(job=next_job, worker_frame=worker_frame)

    def eval(self) -> None:
        self._eval_pool(job=self._job_pool.next_job(), worker_frame=self._env)



"""Unit tests for kms-supported APIs."""

import base64

import boto3
import pytest

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


def create_hmac_key() -> str:
    client = boto3.client("kms", region_name="eu-central-1")

    key_id = client.create_key(
        KeyUsage="GENERATE_VERIFY_MAC", KeySpec="HMAC_512", Policy="My Policy"
    )["KeyMetadata"]["KeyId"]

    return key_id


@mock_aws
def test_generate_mac():
    # Arrange
    key_id = create_hmac_key()
    client = boto3.client("kms", region_name="eu-central-1")

    # Act
    resp = client.generate_mac(
        KeyId=key_id,
        MacAlgorithm="HMAC_SHA_512",
        Message=base64.b64encode("Hello World".encode("utf-8")),
    )

    # Assert
    assert "Mac" in resp
    assert resp["KeyId"] == key_id
    assert resp["MacAlgorithm"] == "HMAC_SHA_512"


@mock_aws
def test_generate_fails_for_non_existent_key():
    # Arrange
    client = boto3.client("kms", region_name="eu-central-1")

    # Act + Assert
    with pytest.raises(client.exceptions.NotFoundException):
        client.generate_mac(
            KeyId="some-key",
            MacAlgorithm="HMAC_SHA_512",
            Message=base64.b64encode("Hello World".encode("utf-8")),
        )


@mock_aws
def test_generate_fails_for_invalid_key_usage():
    # Arrange
    client = boto3.client("kms", region_name="eu-central-1")
    key_id = client.create_key(
        KeyUsage="ENCRYPT_DECRYPT", KeySpec="HMAC_512", Policy="My Policy"
    )["KeyMetadata"]["KeyId"]

    # Act + Assert
    with pytest.raises(client.exceptions.InvalidKeyUsageException):
        client.generate_mac(
            KeyId=key_id,
            MacAlgorithm="HMAC_SHA_512",
            Message=base64.b64encode("Hello World".encode("utf-8")),
        )


@mock_aws
def test_generate_fails_for_invalid_key_spec():
    # Arrange
    client = boto3.client("kms", region_name="eu-central-1")
    key_id = client.create_key(
        KeyUsage="GENERATE_VERIFY_MAC", KeySpec="RSA_2048", Policy="My Policy"
    )["KeyMetadata"]["KeyId"]

    # Act + Assert
    with pytest.raises(client.exceptions.InvalidKeyUsageException):
        client.generate_mac(
            KeyId=key_id,
            MacAlgorithm="HMAC_SHA_512",
            Message=base64.b64encode("Hello World".encode("utf-8")),
        )


@mock_aws
def test_verify_mac():
    # Arrange
    key_id = create_hmac_key()
    client = boto3.client("kms", region_name="eu-central-1")
    mac = client.generate_mac(
        KeyId=key_id,
        MacAlgorithm="HMAC_SHA_512",
        Message=base64.b64encode("Hello World".encode("utf-8")),
    )["Mac"]

    # Act
    resp = client.verify_mac(
        KeyId=key_id,
        MacAlgorithm="HMAC_SHA_512",
        Message=base64.b64encode("Hello World".encode("utf-8")),
        Mac=mac,
    )

    # Assert
    assert resp["KeyId"] == key_id
    assert resp["MacAlgorithm"] == "HMAC_SHA_512"
    assert resp["MacValid"] is True


@mock_aws
def test_verify_mac_fails_for_another_key_id():
    # Arrange
    key_id = create_hmac_key()
    other_key_id = create_hmac_key()
    client = boto3.client("kms", region_name="eu-central-1")
    mac = client.generate_mac(
        KeyId=key_id,
        MacAlgorithm="HMAC_SHA_512",
        Message=base64.b64encode("Hello World".encode("utf-8")),
    )["Mac"]

    # Act + Assert
    with pytest.raises(client.exceptions.KMSInvalidMacException):
        client.verify_mac(
            KeyId=other_key_id,
            MacAlgorithm="HMAC_SHA_512",
            Message=base64.b64encode("Hello World".encode("utf-8")),
            Mac=mac,
        )



"""s3tables base URL and path."""

from .responses import S3TablesResponse

url_bases = [
    r"https?://s3tables\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/buckets$": S3TablesResponse.dispatch,
    "{0}/buckets/(?P<tableBucketARN>.+)$": S3TablesResponse.dispatch,
    "{0}/buckets/(?P<tableBucketARN_pt_1>[^/]+)/(?P<tableBucketARN_pt_2>[^/]+)$": S3TablesResponse.dispatch,
    "{0}/namespaces/(?P<tableBucketARN>.+)$": S3TablesResponse.dispatch,
    "{0}/tables/(?P<tableBucketARN>[^/]+)/(?P<namespace>[^/]+)$": S3TablesResponse.dispatch,
    "{0}/tables/(?P<tableBucketARN>[^/]+)/(?P<namespace>[^/]+)/(?P<name>[^/]+)$": S3TablesResponse.dispatch,
    "{0}/tables/(?P<tableBucketARN>[^/]+)$": S3TablesResponse.dispatch,
    "{0}/tables/(?P<tableBucketARN_pt_1>[^/]+)/(?P<tableBucketARN_pt_2>[^/]+)/(?P<namespace>[^/]+)/(?P<name>[^/]+)$": S3TablesResponse.dispatch,
    "{0}/tables/(?P<tableBucketARN>[^/]+)/(?P<namespace>[^/]+)/(?P<name>[^/]+)/metadata-location$": S3TablesResponse.dispatch,
    "{0}/tables/(?P<tableBucketARN_pt_1>[^/]+)/(?P<tableBucketARN_pt_2>[^/]+)/(?P<namespace>[^/]+)/(?P<name>[^/]+)/metadata-location$": S3TablesResponse.dispatch,
    "{0}/tables/(?P<tableBucketARN>[^/]+)/(?P<namespace>[^/]+)/(?P<name>[^/]+)/rename$": S3TablesResponse.dispatch,
    "{0}/tables/(?P<tableBucketARN_pt_1>[^/]+)/(?P<tableBucketARN_pt_2>[^/]+)/(?P<namespace>[^/]+)/(?P<name>[^/]+)/rename$": S3TablesResponse.dispatch,
}



from __future__ import annotations

from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value import (
    JSONataTemplateValue,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class JSONataTemplateBinding(EvalComponent):
    identifier: str
    value: JSONataTemplateValue

    def __init__(self, identifier: str, value: JSONataTemplateValue):
        self.identifier = identifier
        self.value = value

    def _eval_body(self, env: Environment) -> None:
        binding_container: dict = env.stack.pop()
        self.value.eval(env=env)
        value = env.stack.pop()
        binding_container[self.identifier] = value
        env.stack.append(binding_container)



import json
import logging
from typing import Dict, Optional, Set, Tuple

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task import (
    lambda_eval_utils,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    ComputedCredentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceCondition,
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails

LOG = logging.getLogger(__name__)


_SUPPORTED_INTEGRATION_PATTERNS: Set[ResourceCondition] = {
    ResourceCondition.WaitForTaskToken,
}
_SUPPORTED_API_PARAM_BINDINGS: Dict[str, Set[str]] = {
    "invoke": {
        "ClientContext",
        "FunctionName",
        "Qualifier",
        "Payload",
        # Outside the specification, but supported in practice:
        "LogType",
    }
}


class StateTaskServiceLambda(StateTaskServiceCallback):
    def __init__(self):
        super().__init__(supported_integration_patterns=_SUPPORTED_INTEGRATION_PATTERNS)

    def _get_supported_parameters(self) -> Optional[Set[str]]:
        return _SUPPORTED_API_PARAM_BINDINGS.get(self.resource.api_action.lower())

    @staticmethod
    def _error_cause_from_client_error(client_error: ClientError) -> Tuple[str, str]:
        error_code: str = client_error.response["Error"]["Code"]
        error_msg: str = client_error.response["Error"]["Message"]
        response_details = "; ".join(
            [
                "Service: AWSLambda",
                f"Status Code: {client_error.response['ResponseMetadata']['HTTPStatusCode']}",
                f"Error Code: {error_code}",
                f"Request ID: {client_error.response['ResponseMetadata']['RequestId']}",
                "Proxy: null",
            ]
        )
        error = f"Lambda.{error_code}"
        cause = f"{error_msg} ({response_details})"
        return error, cause

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, lambda_eval_utils.LambdaFunctionErrorException):
            cause = ex.payload
            try:
                cause_object = json.loads(cause)
                error = cause_object["errorType"]
            except Exception as ex:
                LOG.warning(
                    "Could not retrieve 'errorType' field from LambdaFunctionErrorException object: %s",
                    ex,
                )
                error = "Exception"
            error_name = CustomErrorName(error)
        elif isinstance(ex, ClientError):
            error, cause = self._error_cause_from_client_error(ex)
            error_name = CustomErrorName(error)
        else:
            return super()._from_error(env=env, ex=ex)
        return FailureEvent(
            env=env,
            error_name=error_name,
            event_type=HistoryEventType.TaskFailed,
            event_details=EventDetails(
                taskFailedEventDetails=TaskFailedEventDetails(
                    error=error,
                    cause=cause,
                    resource=self._get_sfn_resource(),
                    resourceType=self._get_sfn_resource_type(),
                )
            ),
        )

    def _normalise_parameters(
        self,
        parameters: dict,
        boto_service_name: Optional[str] = None,
        service_action_name: Optional[str] = None,
    ) -> None:
        # Run Payload value casting before normalisation.
        if "Payload" in parameters:
            parameters["Payload"] = lambda_eval_utils.to_payload_type(
                parameters["Payload"]
            )
        super()._normalise_parameters(
            parameters=parameters,
            boto_service_name=boto_service_name,
            service_action_name=service_action_name,
        )

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ):
        lambda_eval_utils.exec_lambda_function(
            env=env,
            parameters=normalised_parameters,
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            credentials=task_credentials,
        )



from typing import Any, Callable, Optional, Set

import boto3
from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import HistoryEventType, TaskFailedEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    ComputedCredentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceCondition,
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str

_SUPPORTED_INTEGRATION_PATTERNS: Set[ResourceCondition] = {
    ResourceCondition.Sync,
}

# Set of JobRunState value that indicate the JobRun had terminated in an abnormal state.
_JOB_RUN_STATE_ABNORMAL_TERMINAL_VALUE: Set[str] = {"FAILED", "TIMEOUT", "ERROR"}

# Set of JobRunState values that indicate the JobRun has terminated.
_JOB_RUN_STATE_TERMINAL_VALUES: Set[str] = {
    "STOPPED",
    "SUCCEEDED",
    *_JOB_RUN_STATE_ABNORMAL_TERMINAL_VALUE,
}

# The handler function name prefix for StateTaskServiceGlue objects.
_HANDLER_REFLECTION_PREFIX: str = "_handle_"
# The sync handler function name prefix for StateTaskServiceGlue objects.
_SYNC_HANDLER_REFLECTION_PREFIX: str = "_sync_to_"
# The type of (sync)handler function for StateTaskServiceGlue objects.
_API_ACTION_HANDLER_TYPE = Callable[
    [Environment, ResourceRuntimePart, dict, ComputedCredentials], None
]
# The type of (sync)handler builder function for StateTaskServiceGlue objects.
_API_ACTION_HANDLER_BUILDER_TYPE = Callable[
    [Environment, ResourceRuntimePart, dict, ComputedCredentials],
    Callable[[], Optional[Any]],
]


class StateTaskServiceGlue(StateTaskServiceCallback):
    def __init__(self):
        super().__init__(supported_integration_patterns=_SUPPORTED_INTEGRATION_PATTERNS)

    def _get_api_action_handler(self) -> _API_ACTION_HANDLER_TYPE:
        api_action = self._get_boto_service_action()
        handler_name = _HANDLER_REFLECTION_PREFIX + api_action
        resolver_handler = getattr(self, handler_name)
        if resolver_handler is None:
            raise ValueError(f"Unknown or unsupported glue action '{api_action}'.")
        return resolver_handler

    def _get_api_action_sync_builder_handler(self) -> _API_ACTION_HANDLER_BUILDER_TYPE:
        api_action = self._get_boto_service_action()
        handler_name = _SYNC_HANDLER_REFLECTION_PREFIX + api_action
        resolver_handler = getattr(self, handler_name)
        if resolver_handler is None:
            raise ValueError(f"Unknown or unsupported glue action '{api_action}'.")
        return resolver_handler

    @staticmethod
    def _get_glue_client(
        resource_runtime_part: ResourceRuntimePart,
        task_credentials: ComputedCredentials,
    ) -> boto3.client:
        return boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service="glue",
            credentials=task_credentials,
        )

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, ClientError):
            error_code = ex.response["Error"]["Code"]
            error_name: str = f"Glue.{error_code}"
            return FailureEvent(
                env=env,
                error_name=CustomErrorName(error_name),
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    taskFailedEventDetails=TaskFailedEventDetails(
                        error=error_name,
                        cause=ex.response["Error"]["Message"],
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                    )
                ),
            )
        return super()._from_error(env=env, ex=ex)

    def _wait_for_task_token(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
    ) -> None:
        raise RuntimeError(
            f"Unsupported .waitForTaskToken callback procedure in resource {self.resource.resource_arn}"
        )

    def _handle_start_job_run(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        computed_credentials: ComputedCredentials,
    ):
        glue_client = self._get_glue_client(
            resource_runtime_part=resource_runtime_part,
            task_credentials=computed_credentials,
        )
        response = glue_client.start_job_run(**normalised_parameters)
        response.pop("ResponseMetadata", None)
        # AWS StepFunctions extracts the JobName from the request and inserts it into the response, which
        # normally only contains JobRunID; as this is a required field for start_job_run, the access at
        # this depth is safe.
        response["JobName"] = normalised_parameters.get("JobName")
        env.stack.append(response)

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ):
        # Source the action handler and delegate the evaluation.
        api_action_handler = self._get_api_action_handler()
        api_action_handler(
            env, resource_runtime_part, normalised_parameters, task_credentials
        )

    def _sync_to_start_job_run(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> Callable[[], Optional[Any]]:
        # Poll the job run state from glue, using GetJobRun until the job has terminated. Hence, append the output
        # of GetJobRun to the state.

        # Access the JobName and the JobRunId from the StartJobRun output call that must
        # have occurred before this point.
        start_job_run_output: dict = env.stack.pop()
        job_name: str = start_job_run_output["JobName"]
        job_run_id: str = start_job_run_output["JobRunId"]

        glue_client = self._get_glue_client(
            resource_runtime_part=resource_runtime_part,
            task_credentials=task_credentials,
        )

        def _sync_resolver() -> Optional[Any]:
            # Sample GetJobRun until completion.
            get_job_run_response: dict = glue_client.get_job_run(
                JobName=job_name, RunId=job_run_id
            )
            job_run: dict = get_job_run_response["JobRun"]
            job_run_state: str = job_run["JobRunState"]

            # If the job run has not terminated, continue and check later.
            is_terminated: bool = job_run_state in _JOB_RUN_STATE_TERMINAL_VALUES
            if not is_terminated:
                return None

            # AWS StepFunctions appears to append attach the JobName to the output both in case of error or success.
            job_run["JobName"] = job_name

            # If the job run terminated in a normal state, return the result.
            is_abnormal_termination = (
                job_run_state in _JOB_RUN_STATE_ABNORMAL_TERMINAL_VALUE
            )
            if not is_abnormal_termination:
                return job_run

            # If the job run has terminated with an abnormal state, raise the error in stepfunctions.
            raise FailureEventException(
                FailureEvent(
                    env=env,
                    error_name=StatesErrorName(
                        typ=StatesErrorNameType.StatesTaskFailed
                    ),
                    event_type=HistoryEventType.TaskFailed,
                    event_details=EventDetails(
                        taskFailedEventDetails=TaskFailedEventDetails(
                            resource=self._get_sfn_resource(),
                            resourceType=self._get_sfn_resource_type(),
                            error=StatesErrorNameType.StatesTaskFailed.to_name(),
                            cause=to_json_str(job_run),
                        )
                    ),
                )
            )

        return _sync_resolver

    def _build_sync_resolver(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> Callable[[], Optional[Any]]:
        sync_resolver_builder = self._get_api_action_sync_builder_handler()
        sync_resolver = sync_resolver_builder(
            env, resource_runtime_part, normalised_parameters, task_credentials
        )
        return sync_resolver



"""Unit tests for sagemaker-supported APIs."""

import time
from datetime import datetime

import boto3
import pytest
from botocore.exceptions import ClientError
from dateutil.tz import tzutc

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html

ACCOUNT_ID = "123456789012"


@mock_aws
def test_create_model_card():
    client = boto3.client("sagemaker", region_name="us-east-1")
    resp = client.create_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model"}}',
    )

    assert (
        resp["ModelCardArn"]
        == f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-card/my-first-model-card"
    )

    with pytest.raises(ClientError) as e:
        resp = client.create_model_card(
            ModelCardName="my-first-model-card",
            ModelCardStatus="Draft",
            Content='{"model_overview": {"model_description": "my first model"}}',
        )
    assert e.value.response["Error"]["Code"] == "ConflictException"
    assert (
        e.value.response["Error"]["Message"]
        == "Modelcard my-first-model-card already exists"
    )


@mock_aws
def test_update_model_card():
    client = boto3.client("sagemaker", region_name="us-east-1")
    client.create_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model"}}',
    )
    client.update_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model ."}}',
    )

    resp = client.list_model_card_versions(ModelCardName="my-first-model-card").get(
        "ModelCardVersionSummaryList"
    )
    assert (len(resp)) == 2

    with pytest.raises(ClientError) as e:
        client.update_model_card(
            ModelCardName="my-second-model-card",
            ModelCardStatus="Draft",
            Content='{"model_overview": {"model_description": "my first model ."}}',
        )
    assert e.value.response["Error"]["Code"] == "ResourceNotFound"
    assert (
        e.value.response["Error"]["Message"]
        == "Modelcard my-second-model-card does not exist."
    )


@mock_aws
def test_model_card_tags():
    client = boto3.client("sagemaker", region_name="us-east-1")
    client.create_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model"}}',
        Tags=[{"Key": "Mytag", "Value": "MyValue"}],
    )

    model_card_arn = (
        f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-card/my-first-model-card"
    )
    tags = client.list_tags(ResourceArn=model_card_arn).get("Tags")
    assert [{"Key": "Mytag", "Value": "MyValue"}] == tags

    client.update_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model >:)"}}',
    )

    model_card_arn = (
        f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-card/my-first-model-card"
    )
    tags = client.list_tags(ResourceArn=model_card_arn).get("Tags")
    assert [{"Key": "Mytag", "Value": "MyValue"}] == tags

    client.add_tags(
        ResourceArn=model_card_arn, Tags=[{"Key": "Mytag2", "Value": "MyValue2"}]
    )
    tags = client.list_tags(ResourceArn=model_card_arn).get("Tags")
    assert {"Key": "Mytag2", "Value": "MyValue2"} in tags

    tags_to_delete = ["Mytag"]
    client.delete_tags(ResourceArn=model_card_arn, TagKeys=tags_to_delete)
    tags = client.list_tags(ResourceArn=model_card_arn).get("Tags")
    assert {"Key": "Mytag", "Value": "MyValue"} not in tags


@mock_aws
def test_list_model_cards_basic():
    client = boto3.client("sagemaker", region_name="us-east-1")

    cards_to_create = ["first", "second", "third"]
    for c in cards_to_create:
        client.create_model_card(
            ModelCardName=f"my-{c}-model-card",
            ModelCardStatus="Draft",
            Content='{"model_overview": {"model_description": f"my {c} model"}}',
        )
    resp = client.list_model_cards().get("ModelCardSummaries")
    assert len(resp) == 3
    for i, r in enumerate(resp):
        assert (
            r["ModelCardName"] == f"my-{cards_to_create[i]}-model-card"
        ), "model card name didn't match expected"
        assert (
            r["ModelCardArn"]
            == f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-card/my-{cards_to_create[i]}-model-card"
        ), "model_card_arn didn't match expected"
        assert (
            r["ModelCardStatus"] == "Draft"
        ), "model card status didn't match expected"


@mock_aws
def test_list_model_cards_advanced():
    client = boto3.client("sagemaker", region_name="us-east-1")

    cards_to_create = ["first", "second", "third"]
    for c in cards_to_create:
        client.create_model_card(
            ModelCardName=f"my-{c}-model-card",
            ModelCardStatus="Draft",
            Content='{"model_overview": {"model_description": f"my {c} model"}}',
        )
    time.sleep(1)
    datetime_now = datetime.now(tzutc())
    client.create_model_card(
        ModelCardName="my-fourth-model-card",
        ModelCardStatus="Approved",
        Content='{"model_overview": {"model_description": "my fourth model"}}',
    )

    resp = client.list_model_cards(NameContains="first").get("ModelCardSummaries")
    assert len(resp) == 1
    assert resp[0].get("ModelCardName") == "my-first-model-card"

    resp = client.list_model_cards().get("ModelCardSummaries")
    assert len(resp) == 4
    assert resp[-1].get("ModelCardName") == "my-fourth-model-card"

    # Ascending by default
    resp = client.list_model_cards(SortOrder="Descending").get("ModelCardSummaries")
    assert len(resp) == 4
    assert resp[0].get("ModelCardName") == "my-fourth-model-card"

    resp = client.list_model_cards(ModelCardStatus="Approved").get("ModelCardSummaries")
    assert len(resp) == 1
    assert resp[0].get("ModelCardName") == "my-fourth-model-card"

    resp = client.list_model_cards(SortBy="Name").get("ModelCardSummaries")
    assert len(resp) == 4
    assert resp[-1].get("ModelCardName") == "my-third-model-card"

    resp = client.list_model_cards(CreationTimeBefore=datetime_now).get(
        "ModelCardSummaries"
    )
    assert len(resp) == 3
    assert resp[-1].get("ModelCardName") == "my-third-model-card"

    resp = client.list_model_cards(CreationTimeAfter=datetime_now).get(
        "ModelCardSummaries"
    )
    assert len(resp) == 1
    assert resp[0].get("ModelCardName") == "my-fourth-model-card"


@mock_aws
def test_list_model_card_versions_basic():
    client = boto3.client("sagemaker", region_name="us-east-1")
    client.create_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model"}}',
    )
    time.sleep(1)
    client.update_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model. :))"}}',
    )
    resp = client.list_model_card_versions(ModelCardName="my-first-model-card").get(
        "ModelCardVersionSummaryList"
    )
    assert len(resp) == 2, f"Expected 2 model card versions, found {len(resp)}"
    assert resp[0].get("ModelCardVersion") == 1
    assert resp[1].get("ModelCardVersion") == 2
    assert resp[0].get("CreationTime") == resp[1].get("CreationTime")
    assert resp[0].get("LastModifiedTime") < resp[1].get("LastModifiedTime")  # type: ignore


@mock_aws
def test_list_model_card_versions_advanced():
    client = boto3.client("sagemaker", region_name="us-east-1")
    client.create_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model"}}',
    )
    client.update_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model. :))"}}',
    )
    client.update_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Approved",
        Content='{"model_overview": {"model_description": "my first model. :))"}}',
    )
    client.create_model_card(
        ModelCardName="my-second-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model. :))"}}',
    )

    resp = client.list_model_card_versions(ModelCardName="my-first-model-card").get(
        "ModelCardVersionSummaryList"
    )
    assert len(resp) == 3
    assert [r.get("ModelCardVersion") for r in resp] == [1, 2, 3]

    resp = client.list_model_card_versions(
        ModelCardName="my-first-model-card", SortOrder="Descending"
    ).get("ModelCardVersionSummaryList")
    assert len(resp) == 3
    assert [r.get("ModelCardVersion") for r in resp] == [3, 2, 1]

    resp = client.list_model_card_versions(
        ModelCardName="my-first-model-card", ModelCardStatus="Approved"
    ).get("ModelCardVersionSummaryList")
    assert len(resp) == 1
    assert [r.get("ModelCardVersion") for r in resp] == [3]

    resp = client.list_model_card_versions(ModelCardName="my-second-model-card").get(
        "ModelCardVersionSummaryList"
    )
    assert len(resp) == 1

    with pytest.raises(ClientError) as e:
        client.list_model_card_versions(ModelCardName="I-dont-exist")
    assert e.value.response["Error"]["Code"] == "ResourceNotFound"


@mock_aws
def test_describe_model_card():
    client = boto3.client("sagemaker", region_name="us-east-1")

    client.create_model_card(
        ModelCardName="my-first-model-card",
        SecurityConfig={"KmsKeyId": "1234abcd-12ab-34cd-56ef-1234567890ab"},
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": "my first model"}}',
    )

    resp = client.describe_model_card(ModelCardName="my-first-model-card")
    assert (
        resp["ModelCardArn"]
        == f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-card/my-first-model-card"
    )
    assert resp["ModelCardName"] == "my-first-model-card"
    assert resp["ModelCardVersion"] == 1
    assert (
        resp["Content"] == '{"model_overview": {"model_description": "my first model"}}'
    )
    assert resp["ModelCardStatus"] == "Draft"
    assert resp["SecurityConfig"] == {
        "KmsKeyId": "1234abcd-12ab-34cd-56ef-1234567890ab"
    }
    assert resp["CreatedBy"] == {}
    assert resp["LastModifiedBy"] == {}

    client.update_model_card(
        ModelCardName="my-first-model-card",
        Content='{"model_overview": {"model_description": "my first model >:)"}}',
        ModelCardStatus="Approved",
    )

    resp = client.describe_model_card(
        ModelCardName="my-first-model-card", ModelCardVersion=1
    )

    assert (
        resp["ModelCardArn"]
        == f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-card/my-first-model-card"
    )
    assert resp["ModelCardName"] == "my-first-model-card"
    assert resp["ModelCardVersion"] == 1
    assert (
        resp["Content"] == '{"model_overview": {"model_description": "my first model"}}'
    )
    assert resp["ModelCardStatus"] == "Draft"
    assert resp["SecurityConfig"] == {
        "KmsKeyId": "1234abcd-12ab-34cd-56ef-1234567890ab"
    }
    assert resp["CreatedBy"] == {}
    assert resp["LastModifiedBy"] == {}

    resp = client.describe_model_card(ModelCardName="my-first-model-card")

    assert (
        resp["ModelCardArn"]
        == f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-card/my-first-model-card"
    )
    assert resp["ModelCardName"] == "my-first-model-card"
    assert resp["ModelCardVersion"] == 2
    assert (
        resp["Content"]
        == '{"model_overview": {"model_description": "my first model >:)"}}'
    )
    assert resp["ModelCardStatus"] == "Approved"
    assert resp["SecurityConfig"] == {
        "KmsKeyId": "1234abcd-12ab-34cd-56ef-1234567890ab"
    }
    assert resp["CreatedBy"] == {}
    assert resp["LastModifiedBy"] == {}

    resp = client.describe_model_card(
        ModelCardName="my-first-model-card", ModelCardVersion=2
    )

    assert (
        resp["ModelCardArn"]
        == f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-card/my-first-model-card"
    )
    assert resp["ModelCardName"] == "my-first-model-card"
    assert resp["ModelCardVersion"] == 2
    assert (
        resp["Content"]
        == '{"model_overview": {"model_description": "my first model >:)"}}'
    )
    assert resp["ModelCardStatus"] == "Approved"
    assert resp["SecurityConfig"] == {
        "KmsKeyId": "1234abcd-12ab-34cd-56ef-1234567890ab"
    }
    assert resp["CreatedBy"] == {}
    assert resp["LastModifiedBy"] == {}

    with pytest.raises(ClientError) as e:
        client.describe_model_card(
            ModelCardName="my-first-model-card", ModelCardVersion=3
        )
    assert e.value.response["Error"]["Code"] == "ResourceNotFound"
    assert (
        e.value.response["Error"]["Message"]
        == "Modelcard with name my-first-model-card and version: 3 does not exist"
    )


@mock_aws
def test_delete_model_card():
    client = boto3.client("sagemaker", region_name="eu-west-1")

    _ = client.create_model_card(
        ModelCardName="my-first-model-card",
        ModelCardStatus="Draft",
        Content='{"model_overview": {"model_description": f"my first model"}}',
    )

    _ = client.delete_model_card(ModelCardName="my-first-model-card")
    resp = client.list_model_cards().get("ModelCardSummaries")
    assert len(resp) == 0

    with pytest.raises(ClientError) as e:
        client.delete_model_card(ModelCardName="my-first-model-card")
    assert e.value.response["Error"]["Code"] == "ResourceNotFound"



"""OpenSearchIngestionBackend class with methods for supported APIs."""

from datetime import datetime
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional

import yaml

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.moto_api._internal import mock_random as random
from moto.moto_api._internal.managed_state_model import ManagedState
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition

from .exceptions import (
    InvalidVPCOptionsException,
    PipelineAlreadyExistsException,
    PipelineInvalidStateException,
    PipelineNotFoundException,
    SecurityGroupNotFoundException,
    SubnetNotFoundException,
)

if TYPE_CHECKING:
    from moto.ec2.models import EC2Backend


class Pipeline(ManagedState, BaseModel):
    CREATING_REASON = "The pipeline is being created. It is not able to ingest data."
    ACTIVE_REASON = "The pipeline is ready to ingest data."
    DELETING_REASON = "The pipeline is being deleted"
    STOPPING_REASON = "The pipeline is being stopped"
    STOPPED_REASON = "The pipeline is stopped"
    STARTING_REASON = "The pipeline is starting. It is not able to ingest data"
    UPDATING_REASON = "An update was triggered for the pipeline. It is still available to ingest data."

    STATUS_REASON_MAP: ClassVar[Dict[str, str]] = {
        "CREATING": CREATING_REASON,
        "ACTIVE": ACTIVE_REASON,
        "STOPPING": STOPPING_REASON,
        "STOPPED": STOPPED_REASON,
        "STARTING": STARTING_REASON,
        "UPDATING": UPDATING_REASON,
        "DELETING": DELETING_REASON,
    }

    def __init__(
        self,
        pipeline_name: str,
        account_id: str,
        region: str,
        min_units: int,
        max_units: int,
        pipeline_configuration_body: str,
        log_publishing_options: Optional[Dict[str, Any]],
        vpc_options: Optional[Dict[str, Any]],
        buffer_options: Optional[Dict[str, Any]],
        encryption_at_rest_options: Optional[Dict[str, Any]],
        ingest_endpoint_urls: List[str],
        serverless: bool,
        vpc_endpoint_service: Optional[str],
        vpc_endpoint: Optional[str],
        vpc_id: Optional[str],
        backend: "OpenSearchIngestionBackend",
    ):
        ManagedState.__init__(
            self,
            model_name="osis::pipeline",
            transitions=[
                ("CREATING", "ACTIVE"),
                ("UPDATING", "ACTIVE"),
                ("DELETING", "DELETED"),
                ("STOPPING", "STOPPED"),
                ("STARTING", "ACTIVE"),
            ],
        )

        self.pipeline_name = pipeline_name
        self.account_id = account_id
        self.region = region
        self.min_units = min_units
        self.max_units = max_units
        self.pipeline_configuration_body_str = pipeline_configuration_body
        self.pipeline_configuration_body = yaml.safe_load(pipeline_configuration_body)
        self.log_publishing_options = log_publishing_options
        self.vpc_options = vpc_options
        self.buffer_options = buffer_options
        self.encryption_at_rest_options = encryption_at_rest_options
        self.ingest_endpoint_urls = ingest_endpoint_urls
        self.serverless = serverless
        self.vpc_endpoint_service = vpc_endpoint_service
        self.vpc_endpoint = vpc_endpoint
        self.vpc_id = vpc_id
        self.backend = backend

        self.status = "CREATING"
        self.arn = self._get_arn(self.pipeline_name)
        self.destinations = self._update_destinations()

        if (
            self.vpc_options is None
            or self.vpc_options.get("VpcEndpointManagement", "SERVICE") == "SERVICE"
        ):
            # Not returned in this case
            self.vpc_endpoint_service = None

        self.service_vpc_endpoints = self._get_service_vpc_endpoints()
        self.created_at: datetime = datetime.now()
        self.last_updated_at: datetime = datetime.now()

    def _get_arn(self, name: str) -> str:
        return f"arn:{get_partition(self.region)}:osis:{self.region}:{self.account_id}:pipeline/{name}"

    def _get_service_vpc_endpoints(self) -> Optional[List[Dict[str, str]]]:
        # ServiceVpcEndpoint.VpcEndpointId not implemented
        if self.serverless:
            return [{"ServiceName": "OPENSEARCH_SERVERLESS"}]
        else:
            return None

    def _update_destinations(self) -> List[Dict[str, str]]:
        destinations = []
        for sub_pipeline in self.pipeline_configuration_body:
            if sub_pipeline != "version":
                for sink in self.pipeline_configuration_body[sub_pipeline]["sink"]:
                    for sink_type, sink_config in sink.items():
                        if sink_type == "opensearch":
                            if sink_config["aws"].get("serverless") is True:
                                service_name = "OpenSearch_Serverless"
                            else:
                                service_name = "OpenSearch"
                            endpoint = sink_config["hosts"][0]
                        elif sink_type == "s3":
                            service_name = "S3"
                            endpoint = sink_config["bucket"]
                        else:
                            continue
                        destinations.append(
                            {"ServiceName": service_name, "Endpoint": endpoint}
                        )
        return destinations

    @staticmethod
    def is_serverless(pipeline_body: Dict[str, Any]) -> bool:  # type: ignore[misc]
        serverless = False
        for sub_pipeline in pipeline_body:
            if sub_pipeline != "version":
                for sink in pipeline_body[sub_pipeline]["sink"]:
                    for _, sink_config in sink.items():
                        serverless = (
                            sink_config.get("aws", {}).get("serverless", False)
                            or serverless
                        )
                source_type = list(pipeline_body[sub_pipeline]["source"].keys())[0]
                source_config = pipeline_body[sub_pipeline]["source"][source_type]
                serverless = (
                    source_config.get("aws", {}).get("serverless", False) or serverless
                )
        return serverless

    def delete(self) -> None:
        self.status = "DELETING"
        self.set_last_updated()

    def get_created_at(self) -> str:
        return self.created_at.astimezone().isoformat()

    def get_last_updated_at(self) -> str:
        return self.last_updated_at.astimezone().isoformat()

    def set_last_updated(self) -> None:
        self.last_updated_at = datetime.now()

    def start(self) -> None:
        self.status = "STARTING"
        self.set_last_updated()

    def stop(self) -> None:
        self.status = "STOPPING"
        self.set_last_updated()

    def update(
        self,
        min_units: Optional[int],
        max_units: Optional[int],
        pipeline_configuration_body: Optional[str],
        log_publishing_options: Optional[Dict[str, Any]],
        buffer_options: Optional[Dict[str, Any]],
        encryption_at_rest_options: Optional[Dict[str, Any]],
    ) -> None:
        if min_units is not None:
            self.min_units = min_units
        if max_units is not None:
            self.max_units = max_units
        if pipeline_configuration_body is not None:
            self.pipeline_configuration_body_str = pipeline_configuration_body
            self.pipeline_configuration_body = yaml.safe_load(
                pipeline_configuration_body
            )
        if log_publishing_options is not None:
            self.log_publishing_options = log_publishing_options
        if buffer_options is not None:
            self.buffer_options = buffer_options
        if encryption_at_rest_options is not None:
            self.encryption_at_rest_options = encryption_at_rest_options
        self.destinations = self._update_destinations()
        self.serverless = self.is_serverless(self.pipeline_configuration_body)
        self.service_vpc_endpoints = self._get_service_vpc_endpoints()
        self.status = "UPDATING"
        self.set_last_updated()

    def to_dict(self) -> Dict[str, Any]:
        return {
            "PipelineName": self.pipeline_name,
            "PipelineArn": self.arn,
            "MinUnits": self.min_units,
            "MaxUnits": self.max_units,
            "Status": self.status,
            "StatusReason": {
                "Description": self.STATUS_REASON_MAP.get(self.status or "", ""),
            },
            "PipelineConfigurationBody": self.pipeline_configuration_body_str,
            "CreatedAt": self.get_created_at(),
            "LastUpdatedAt": self.get_last_updated_at(),
            "IngestEndpointUrls": self.ingest_endpoint_urls,
            "LogPublishingOptions": self.log_publishing_options,
            "VpcEndpoints": None
            if self.vpc_options is None
            else [
                {
                    "VpcEndpointId": self.vpc_endpoint,
                    "VpcId": self.vpc_id,
                    "VpcOptions": self.vpc_options,
                }
            ],
            "BufferOptions": self.buffer_options,
            "EncryptionAtRestOptions": self.encryption_at_rest_options,
            "VpcEndpointService": self.vpc_endpoint_service,
            "ServiceVpcEndpoints": self.service_vpc_endpoints,
            "Destinations": self.destinations,
            "Tags": self.backend.list_tags_for_resource(self.arn)["Tags"],
        }

    def to_short_dict(self) -> Dict[str, Any]:
        return {
            "Status": self.status,
            "StatusReason": {
                "Description": self.STATUS_REASON_MAP.get(self.status or "", ""),
            },
            "PipelineName": self.pipeline_name,
            "PipelineArn": self.arn,
            "MinUnits": self.min_units,
            "MaxUnits": self.max_units,
            "CreatedAt": self.get_created_at(),
            "LastUpdatedAt": self.get_last_updated_at(),
            "Destinations": self.destinations,
            "Tags": self.backend.list_tags_for_resource(self.arn)["Tags"],
        }


class OpenSearchIngestionBackend(BaseBackend):
    """Implementation of OpenSearchIngestion APIs."""

    PAGINATION_MODEL = {
        "list_pipelines": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "PipelineName",
        },
    }

    PIPELINE_DELETE_VALID_STATES = [
        "UPDATE_FAILED",
        "ACTIVE",
        "START_FAILED",
        "STOPPED",
        "CREATE_FAILED",
    ]
    PIPELINE_STOP_VALID_STATES = ["UPDATE_FAILED", "ACTIVE"]
    PIPELINE_START_VALID_STATES = ["START_FAILED", "STOPPED"]
    PIPELINE_UPDATE_VALID_STATES = [
        "UPDATE_FAILED",
        "ACTIVE",
        "START_FAILED",
        "STOPPED",
    ]

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self._pipelines: Dict[str, Pipeline] = dict()
        self.tagger = TaggingService()

    @property
    def ec2_backend(self) -> "EC2Backend":  # type: ignore[misc]
        from moto.ec2 import ec2_backends

        return ec2_backends[self.account_id][self.region_name]

    @property
    def pipelines(self) -> Dict[str, Pipeline]:
        self._pipelines = {
            name: pipeline
            for name, pipeline in self._pipelines.items()
            if pipeline.status != "DELETED"
        }
        return self._pipelines

    def _get_ingest_endpoint_urls(
        self, pipeline_name: str, endpoint_random_string: str
    ) -> List[str]:
        return [
            f"{pipeline_name}-{endpoint_random_string}.{self.region_name}.osis.amazonaws.com"
        ]

    def _get_random_endpoint_string(self) -> str:
        return random.get_random_string(length=26, lower_case=True)

    def _get_vpc_endpoint(
        self, vpc_id: str, vpc_options: Dict[str, Any], service_name: str
    ) -> Optional[str]:
        if vpc_options.get("VpcEndpointManagement", "SERVICE") == "SERVICE":
            service_managed_endpoint = self.ec2_backend.create_vpc_endpoint(
                vpc_id=vpc_id,
                service_name=service_name,
                endpoint_type="Interface",
                security_group_ids=vpc_options.get("SecurityGroupIds"),
                subnet_ids=vpc_options["SubnetIds"],
                private_dns_enabled=False,
                policy_document="OSIS Test Doc",
                route_table_ids=[],
                tags={"OSISManaged": "true"},
            )
            return service_managed_endpoint.id
        else:
            return None

    def _get_vpc_endpoint_service(
        self, pipeline_name: str, endpoint_random_string: str
    ) -> str:
        return f"com.amazonaws.osis.{self.region_name}.{pipeline_name}-{endpoint_random_string}"

    def _validate_and_get_vpc(self, vpc_options: Dict[str, Any]) -> str:
        from moto.ec2.exceptions import InvalidSubnetIdError

        vpc_id = ""
        for subnet_id in vpc_options["SubnetIds"]:
            try:
                subnet = self.ec2_backend.get_subnet(subnet_id)
            except InvalidSubnetIdError:
                # re-raising for more accurate error message
                raise SubnetNotFoundException(subnet_id)
            if vpc_id == "":
                vpc_id = subnet.vpc_id
            else:
                if subnet.vpc_id != vpc_id:
                    raise InvalidVPCOptionsException(
                        "All specified subnets must belong to the same VPC."
                    )

        for sg_id in vpc_options["SecurityGroupIds"]:
            sg = self.ec2_backend.get_security_group_from_id(sg_id)
            if sg is None:
                raise SecurityGroupNotFoundException(sg_id)

        return vpc_id

    def create_pipeline(
        self,
        pipeline_name: str,
        min_units: int,
        max_units: int,
        pipeline_configuration_body: str,
        log_publishing_options: Optional[Dict[str, Any]],
        vpc_options: Optional[Dict[str, Any]],
        buffer_options: Optional[Dict[str, bool]],
        encryption_at_rest_options: Optional[Dict[str, Any]],
        tags: List[Dict[str, str]],
    ) -> Pipeline:
        if pipeline_name in self.pipelines:
            raise PipelineAlreadyExistsException(pipeline_name)

        serverless = Pipeline.is_serverless(yaml.safe_load(pipeline_configuration_body))

        endpoint_random_string = self._get_random_endpoint_string()
        endpoint_service = self._get_vpc_endpoint_service(
            pipeline_name, endpoint_random_string
        )

        ingestion_endpoint_urls = self._get_ingest_endpoint_urls(
            pipeline_name, endpoint_random_string
        )

        vpc_endpoint = None
        vpc_id = None
        if vpc_options is not None:
            vpc_id = self._validate_and_get_vpc(vpc_options)
            vpc_endpoint = self._get_vpc_endpoint(vpc_id, vpc_options, endpoint_service)

        pipeline = Pipeline(
            pipeline_name,
            self.account_id,
            self.region_name,
            min_units,
            max_units,
            pipeline_configuration_body,
            log_publishing_options,
            vpc_options,
            buffer_options,
            encryption_at_rest_options,
            ingestion_endpoint_urls,
            serverless,
            endpoint_service,
            vpc_endpoint,
            vpc_id,
            backend=self,
        )
        self.pipelines[pipeline_name] = pipeline
        self.tag_resource(pipeline.arn, tags)
        return pipeline

    def delete_pipeline(self, pipeline_name: str) -> None:
        if pipeline_name not in self.pipelines:
            raise PipelineNotFoundException(pipeline_name)
        pipeline = self.pipelines[pipeline_name]
        if pipeline.status not in self.PIPELINE_DELETE_VALID_STATES:
            raise PipelineInvalidStateException(
                "deletion", self.PIPELINE_DELETE_VALID_STATES, pipeline.status
            )
        pipeline.delete()

    def start_pipeline(self, pipeline_name: str) -> Pipeline:
        if pipeline_name not in self.pipelines:
            raise PipelineNotFoundException(pipeline_name)
        pipeline = self.pipelines[pipeline_name]
        if pipeline.status not in self.PIPELINE_START_VALID_STATES:
            raise PipelineInvalidStateException(
                "starting", self.PIPELINE_START_VALID_STATES, pipeline.status
            )
        pipeline.start()
        return pipeline

    def stop_pipeline(self, pipeline_name: str) -> Pipeline:
        if pipeline_name not in self.pipelines:
            raise PipelineNotFoundException(pipeline_name)
        pipeline = self.pipelines[pipeline_name]
        if pipeline.status not in self.PIPELINE_STOP_VALID_STATES:
            raise PipelineInvalidStateException(
                "stopping", self.PIPELINE_STOP_VALID_STATES, pipeline.status
            )
        pipeline.stop()
        return pipeline

    def get_pipeline(self, pipeline_name: str) -> Pipeline:
        if pipeline_name not in self.pipelines:
            raise PipelineNotFoundException(pipeline_name)
        pipeline = self.pipelines[pipeline_name]
        pipeline.advance()
        return pipeline

    @paginate(pagination_model=PAGINATION_MODEL)  # type: ignore
    def list_pipelines(self) -> List[Pipeline]:
        for pipeline in self.pipelines.values():
            pipeline.advance()
        return [p for p in self.pipelines.values()]

    def list_tags_for_resource(self, arn: str) -> Dict[str, List[Dict[str, str]]]:
        return self.tagger.list_tags_for_resource(arn)

    def update_pipeline(
        self,
        pipeline_name: str,
        min_units: Optional[int],
        max_units: Optional[int],
        pipeline_configuration_body: Optional[str],
        log_publishing_options: Optional[Dict[str, Any]],
        buffer_options: Optional[Dict[str, Any]],
        encryption_at_rest_options: Optional[Dict[str, Any]],
    ) -> Pipeline:
        if pipeline_name not in self.pipelines:
            raise PipelineNotFoundException(pipeline_name)
        pipeline = self.pipelines[pipeline_name]
        if pipeline.status not in self.PIPELINE_UPDATE_VALID_STATES:
            raise PipelineInvalidStateException(
                "updates", self.PIPELINE_UPDATE_VALID_STATES, pipeline.status
            )
        pipeline.update(
            min_units,
            max_units,
            pipeline_configuration_body,
            log_publishing_options,
            buffer_options,
            encryption_at_rest_options,
        )
        return pipeline

    def tag_resource(self, arn: str, tags: List[Dict[str, str]]) -> None:
        self.tagger.tag_resource(arn, tags)

    def untag_resource(self, arn: str, tag_keys: List[str]) -> None:
        self.tagger.untag_resource_using_names(arn, tag_keys)


osis_backends = BackendDict(OpenSearchIngestionBackend, "osis")
